/* See the file Distribution for distribution terms.
	(c) Copyright 1994 Ari Halberstadt */

/*	This file contains a popup CDEF that uses the file PopupLib.c to handle
	the details of popup menus. This file contains glue to link the Control
	Manager and the routines in PopupLib.c.

	950826 aih	- popup glue for debugging works with powerpc
	950630 aih	- moved GetMenuHandle before GetResource
	950621 aih	- converted to mwerks
	941226 aih	- checks menuProc field when loading menu
	941126 aih	- automatically adjusts control's min, max
	940706 aih	- adapted for universal headers
	940315 aih	- added check against kPopupVersion before a CDEF is attached during
					debugging with PopupCDEFAttach.
					- Moved description to documentation file.
					- Uses bullet instead of check mark as mark character when using
					window font.
					- Added support for the popupFixedWidth variation code.
	940119 aih	- Added call to GetMenuHandle to allow use of menus created by
					the application that are not in the resource file. Also fixed
					use of an already disposed handle in the CDEF's dispose routine.
					Thanks to Eric Bowman (bobo@reed.edu) for both of these.
	931231 aih	- Returns 1 as part code (same as 7.0 CDEF) instead of inCheckBox
	931226 aih	- Major overhaul. Now supports most of the features of the System 7.0 CDEF.
					- Added functions for installing a glue handle to the popup CDEF,
					allowing the CDEF to be debugged from within an application. The only
					message that can't be debugged this way is the initCntl message.
	931224 aih	- Adapted to use PopupHandle type
					- Removed dependence on other libraries
					- Simplified by removing complicated variation setting code
					based on parsing the title string
	910315 aih	- Fixed bug which caused a crash if the menu resource specified in the
					control's refCon field couldn't be loaded
	910304 aih	- created this file */

#include <limits.h>
#include <Controls.h>
#include <LowMem.h>
#include <Memory.h>
#include <Resources.h>
#include <TextEdit.h>
#include <ToolUtils.h>
#include <Windows.h>
#include "PopupLib.h"

#if __MWERKS__
	#include <A4Stuff.h>
#elif THINK_C
	#include <SetUpA4.h>
#endif

/* mask for the low 31 bits of the calcCRgns message to CDEFs */
#define calcCRgnsMask		(0x7fffffffL)

/* hilite value for a disabled control */
#define kControlDisabled	(255)

/* System software version needed to use this CDEF. Tested with systems
	6.0.5, 7.0, and 7.1.1, and 7.5, but may work on systems as early
	as 4.0.1. */
#define kSystemVersion		(0x0605)

/* The checkMark character is only present in the Chicago system font,
	whereas the bullet character ('') is present in most fonts. When the
	popupUseWFont variation code is used, we change the mark character
	from the default checkMark to kBulletMark. */
#define kBulletMark			('')

/*	When debugging, it's convenient to have the CDEF ignore the
	initCntl message, so that we can send it our own message
	to initialize it when we call PopupCDEFAttach. */
#if DEBUG_CDEF
	#define initCntlMsg		(1234)
#else /* DEBUG_CDEF */
	#define initCntlMsg		initCntl
#endif /* DEBUG_CDEF */

/* draw the control */
static void Draw(ControlHandle ctl, PopupHandle popup, short param)
{
	#pragma unused(param)
	Str255 title;
	Rect bounds;
	
	GetControlTitle(ctl, title);
	bounds = (**ctl).contrlRect;
	PopupDrawSet(popup, false);
	PopupTitleSet(popup, title);
	PopupBoundsSet(popup, &bounds);
	PopupCurrentSet(popup, (**ctl).contrlValue);
	PopupEnableSet(popup, (**ctl).contrlHilite != kControlDisabled);
	PopupDrawSet(popup, true);
	PopupCalculate(popup);
	if ((**ctl).contrlVis)
		PopupDraw(popup);
}

/* return the part of the control that the point is in */
static long Test(ControlHandle ctl, PopupHandle popup, Point where)
{
	#pragma unused(ctl)
	return(PopupWithin(popup, where) ? kPopupPartCode : 0);
}

/* calculate the region containing the control */
static void Calculate(ControlHandle ctl, PopupHandle popup, RgnHandle rgn)
{
	#pragma unused(ctl)
	Rect bounds;
	
	PopupBounds(popup, &bounds);
	RectRgn(rgn, &bounds);
}

/* load the menu */
static MenuHandle LoadMenu(short id, Boolean *createdMenu)
{
	MenuHandle	menu;		/* the menu */
	ResErrUPP	errproc;	/* for saving and restoring ResErrProc */
	short			load;		/* for saving and restoring ResLoad */
	
	*createdMenu = false;

	/* To allow use of menus created by the application that are not in the
		resource file we look for the menu in the menu list. */
	menu = GetMenuHandle(id);
	if (! menu) {

		/* We can tell if the menu hasn't been loaded by a call to GetMenu
			since the menu handle will be a nil resource handle. */
		load = LMGetResLoad();
		LMSetResLoad(false);
		menu = (MenuHandle) GetResource('MENU', id);
		LMSetResLoad(load);
		if (menu && (! *menu || ! (**menu).menuProc)) {

			/* menu hasn't been loaded yet, so we have to call GetMenu */
			errproc = LMGetResErrProc();
			LMSetResErrProc(NULL);
			menu = GetMenu(id);
			LMSetResErrProc(errproc);
			*createdMenu = true;
		}
	}
	return(menu);
}

/* initialize the control */
static void Initialize(ControlHandle ctl, short var)
{
	MenuHandle	menu;				/* the menu */
	PopupHandle popup;			/* the popup menu */
	Rect			bounds;			/* control's bounding rectangle */
	short			nitems;			/* number of items in menu */
	Boolean		createdMenu;	/* true if we created the menu by calling GetMenu */
	
	menu = LoadMenu((**ctl).contrlMin, &createdMenu);
	if (menu) {
		if ((var & popupUseAddResMenu) != 0)
			AppendResMenu(menu, (**ctl).contrlRfCon);
		bounds = (**ctl).contrlRect;
		popup = PopupBegin((GrafPtr) GetWindowPort((**ctl).contrlOwner), menu, &bounds, ctl);
		if (popup) {
			(**popup).state.createdMenu = createdMenu;
			PopupDrawSet(popup, false);
			PopupTitleWidthSet(popup, (**ctl).contrlMax);
			if (((**ctl).contrlValue & popupTitleNoStyle) == 0)
				PopupTitleStyleSet(popup, (**ctl).contrlValue >> CHAR_BIT);
			if (((**ctl).contrlValue & popupTitleRightJust) == popupTitleRightJust)
				PopupJustSet(popup, teFlushRight);
			else if (((**ctl).contrlValue & popupTitleCenterJust) == popupTitleCenterJust)
				PopupJustSet(popup, teCenter);
			if ((var & popupUseWFont) != 0)
				PopupMarkSet(popup, kBulletMark);
			PopupTypeInSet(popup, (var & popupTypeIn) != 0);
			PopupFixedWidthSet(popup, (var & popupFixedWidth) != 0);
			PopupWindowFontSet(popup, (var & popupUseWFont) != 0);
			PopupDrawSet(popup, true);
			PopupCalculate(popup);
			nitems = CountMItems(menu);
			(**ctl).contrlMax = nitems;
			(**ctl).contrlMin = (nitems > 0 ? 1 : 0);
			(**ctl).contrlValue = (nitems > 0 ? 1 : 0);
			(**ctl).contrlData = (Handle) popup;
			(**ctl).contrlAction = (ControlActionUPP) -1L;
		}
	}
}

/* dispose of the control */
static void Dispose(ControlHandle ctl, PopupHandle popup)
{
	MenuHandle menu;
	Boolean createdMenu;
	
	menu = (**popup).menu;
	createdMenu = (**popup).state.createdMenu;
	PopupEnd(popup);
	if (createdMenu)
		ReleaseResource((Handle) menu);
	(**ctl).contrlData = NULL;
}

/* track a mouse click in the control */
static long Track(ControlHandle ctl, PopupHandle popup)
{
	short value;
	short part;

	part = PopupSelect(popup);
	value = PopupCurrent(popup);
	(**ctl).contrlValue = value;
	return(part);
}

/* entry point for CDEF */
pascal long PopupCDEF(short var, ControlHandle ctl, short msg, long param)
{
	PopupHandle	popup = NULL;
	SysEnvRec world;
	long result = 0;
	short nitems;

	/* setup global variable base register */
	#if CDEF
		#if __MWERKS__
			long oldA4 = SetCurrentA4();
		#elif THINK_C
			RememberA0();
			SetUpA4();
		#endif
	#endif
	
	/* check system software */
	(void) SysEnvirons(curSysEnvVers, &world);
	if (world.systemVersion >= kSystemVersion) {
		
		/* execute message */
		popup = (PopupHandle) (**ctl).contrlData;
		if (msg == initCntlMsg) {
			Initialize(ctl, var);
			popup = (PopupHandle) (**ctl).contrlData;
		}
		else if (popup) {
			nitems = CountMItems((**popup).menu);
			(**ctl).contrlMax = nitems;
			(**ctl).contrlMin = (nitems > 0 ? 1 : 0);
			if ((**ctl).contrlValue > (**ctl).contrlMax)
				(**ctl).contrlValue = (**ctl).contrlMax;
			if ((**ctl).contrlValue < (**ctl).contrlMin)
				(**ctl).contrlValue = (**ctl).contrlMin;
			switch (msg) {
			case drawCntl:
				param = LoWord(param); /* see TN196 */
				Draw(ctl, popup, param);
				break;
			case testCntl:
				result = Test(ctl, popup, *(Point *) &param);
				break;
			case calcCRgns:
				param &= calcCRgnsMask;
				/* no break */
			case calcCntlRgn:
			case calcThumbRgn:
				Calculate(ctl, popup, (RgnHandle) param);
				break;
			case dispCntl:
				Dispose(ctl, popup);
				popup = NULL;
				break;
			case autoTrack:

				/*
					Some posters to Usenet have requested a means to determine if no item was
					selected from a popup menu. I have tried to address this request by adding
					the part code kPopupPartNotSelected, which is returned when no selection is
					made from a menu. However, I cannot seem to get TrackControl to return this
					result to the application. The user should therefore check the value of the
					control before and after the call to TrackControl to determine if a different
					item was selected. Alternatively, the user can check the low word of the
					low-memory global MenuDisable. If the low-word of MenuDisable is zero, then
					no item was selected from the popup menu.

					From TechNote TB31:

					_TrackControl
					When a user clicks on your control, you normally call _TrackControl, which is
					supposed to return zero if the user does not change the controls setting or
					the part code if the user does change the setting.  For 'CDEF' resources that
					implement custom dragging, _TrackControl returns zero whether or not the user
					changes the controls setting.  To work around this problem, you must use
					another method to find out if the user has changed the controls setting, such
					as comparing the controls value before and after the call to _TrackControl.

					It's not 100% clear to me what this note is saying. Do they mean a custom
					dragging routine like dragCntl, or actually (and most likely) a custom tracking
					routine? Also, I observed (in System 7.5 on a PowerPC) that TrackControl actually
					sends the control a testCntl message before sending it the autoTrack message.
					Whatever value is returned by the testCntl message is returned as the result of
					TrackControl, regardless of the value returned by the autoTrack message. */

				param = LoWord(param); /* see TN196 */
				result = Track(ctl, popup);
				break;
			}
		}
	}

	/* restore global variable base register */
	#if CDEF
		#if __MWERKS__
			SetA4(oldA4);
		#elif THINK_C
			RestoreA4();
		#endif
	#endif

	return(result);
}

#if CDEF

pascal long main(short var, ControlHandle ctl, short msg, long param)
{
	return(PopupCDEF(var, ctl, msg, param));
}

#else /* CDEF */

/*	Functions for attaching a glue handle so the CDEF can be debugged
	from within an application. */

/* Structure installed in the contrlDefProc field. */
typedef struct {
	PopupGlueUnion glue;		/* glue stuff */
	Handle contrlDefProc;	/* saved value of contrlDefProc field */
} TestGlueType, *TestGluePtr, **TestGlueHandle;

/* Set the control's contrlDefProc field to a small glue handle that will
	jump to PopupCDEF. This makes debugging a lot easier, since you
	can then step through the code with a debugger. */
void PopupCDEFAttach(ControlHandle ctl)
{
	TestGlueHandle binder;
	
	if (! (**ctl).contrlData ||
		 PopupVersion((PopupHandle) (**ctl).contrlData) == kPopupVersion)
	{
		binder = (TestGlueHandle) NewHandleClear(sizeof(TestGlueType));
		if (binder) {
			HLock((Handle) binder);
			PopupGlueInit(&(**binder).glue,
				(ProcPtr) PopupCDEF,
				uppControlDefProcInfo);
			HUnlock((Handle) binder);
			(**binder).contrlDefProc = (**ctl).contrlDefProc;
			(**ctl).contrlDefProc = (Handle) binder;
			if (! (**ctl).contrlData)
				(void) PopupCDEF(GetControlVariant(ctl), ctl, initCntlMsg, 0);
		}
	}
}

/* Dispose of the glue handle created with PopupCDEFAttach and set the
	control's contrlDefProc field to point to its original value. */
void PopupCDEFDetach(ControlHandle ctl)
{
	TestGlueHandle binder;
	
	if (PopupVersion((PopupHandle) (**ctl).contrlData) == kPopupVersion) {
		binder = (TestGlueHandle) (**ctl).contrlDefProc;
		(**ctl).contrlDefProc = (**binder).contrlDefProc;
		DisposeHandle((Handle) binder);
	}
}

#endif /* CDEF */
