/****************************************************************************/
/**                                                                        **/
/**  This file contains the functions necessary to define, draw and        **/
/**  manipulate a pair of robot arms.                                      **/
/**                                                                        **/
/****************************************************************************/

#include <stdio.h>
#include <math.h>
#include <gl/gl.h>
#include "display_list.h"
#include "unit_objects.h"
#include "materials.h"
#include "world.h"
#include "connect.h"
#include "chips.h"
#include "board.h"

#define PI                       3.141592654
#define ARM_PARTS                9
#define DISTANCE_FROM_CENTRE     8.0
#define DISTANCE_ABOVE_BOARD     1.5
#define OPEN_ANGLE               80.0
#define DEGREES_PER_SLOW_FRAME   1.0
#define DEGREES_PER_MEDIUM_FRAME 2.5
#define DEGREES_PER_FAST_FRAME   4.0


enum { X, Y, Z };

enum {
    BASE,
    BASE_PIVOT,
    LOWER_ARM,
    ELBOW,
    UPPER_ARM,
    WRIST,
    LOWER_FINGER,
    UPPER_FINGER,
    CLUTCHES
};

/* Define the sizes of the various arm parts. */

static float scales[ARM_PARTS][3]  = {
    /* X    Y    Z */
    { 6.0, 1.8, 6.0 },  /* BASE */
    { 3.0, 1.2, 3.0 },  /* BASE_PIVOT */
    { 7.0, 0.7, 0.7 },  /* LOWER_ARM */
    { 0.8, 0.8, 0.8 },  /* ELBOW */
    { 6.0, 0.6, 0.6 },  /* UPPER_ARM */
    { 0.7, 1.4, 0.7 },  /* WRIST */
    { 2.0, 0.5, 0.5 },  /* LOWER_FINGER */
    { 0.5, 2.0, 0.5 },  /* UPPER_FINGER */
    { 0.0, 0.0, 0.0 }   /* CLUTCHES (unused) */
};

static float angles[2][ARM_PARTS];

/* Define the default angle of each of the joints of each of the arms. */

static float default_angles[2][ARM_PARTS]  = {
    {
        0.0,            /* BASE (unused) */
        0.0,            /* BASE_PIVOT */
        120.0,          /* LOWER_ARM */
        -110.0,         /* ELBOW */
        0.0,            /* UPPER_ARM */
        -30.0,          /* WRIST */
        0.0,            /* LOWER_FINGER (unused) */
        0.0,            /* UPPER_FINGER (unused) */
        45.0            /* CLUTCHES */
    },
    {
        0.0,            /* BASE (unused) */
        180.0,          /* BASE_PIVOT */
        110.0,          /* LOWER_ARM */
        -120.0,         /* ELBOW */
        0.0,            /* UPPER_ARM */
        20.0,           /* WRIST */
        0.0,            /* LOWER_FINGER (unused) */
        0.0,            /* UPPER_FINGER (unused) */
        45.0            /* CLUTCHES */
    }
};

/* Define a neutral angle of each of the joints of each of the arms. */
/* These neutral angles define an intermediate arm configuration     */
/* which each robot arm will move to as an intermediate stage in     */
/* moving a chip from a pile to the game board.  This neutral step   */
/* is necessary to avoid having the arm sweep into the game board.   */

static float neutral_angles[2][ARM_PARTS]  = {
    {
        0.0,            /* BASE (unused) */
        30.0,           /* BASE_PIVOT */
        100.0,          /* LOWER_ARM */
        -80.0,          /* ELBOW */
        0.0,            /* UPPER_ARM */
        -60.0,          /* WRIST */
        0.0,            /* LOWER_FINGER (unused) */
        0.0,            /* UPPER_FINGER (unused) */
        45.0            /* CLUTCHES */
    },
    {
        0.0,            /* BASE (unused) */
        210.0,          /* BASE_PIVOT */
        100.0,          /* LOWER_ARM */
        -80.0,          /* ELBOW */
        0.0,            /* UPPER_ARM */
        -60.0,          /* WRIST */
        0.0,            /* LOWER_FINGER (unused) */
        0.0,            /* UPPER_FINGER (unused) */
        45.0            /* CLUTCHES */
    }
};

Display_list robot_arm[2];

/* Define some variables local to this file. */

static float degrees_per_frame = DEGREES_PER_FAST_FRAME;
static float chip_translate[3];
static float flip_angle = 180.0, square_angle = 90.0;
static Object cube, coarse_disk, medium_disk, fine_disk, nail, chip_object;
static Object chip[2] = { -1, -1 };
static float clutch_size;
static int arm_materials[2] = { RED_ARM_MATERIAL, BLUE_ARM_MATERIAL };
static int chip_materials[2] = { RED_CHIP_MATERIAL, BLUE_CHIP_MATERIAL };

static Display_list
    upper_finger[2],
    lower_finger[2],
    finger[2],
    first_finger[2],
    second_finger[2],
    wrist[2],
    current_chip[2],
    clutches[2],
    elbow[2],
    upper_arm[2],
    all_above_lower[2],
    lower_arm[2],
    all_above_pivot[2],
    base_pivot[2],
    all_but_base[2],
    base[2];

/* Define some functions local to this file. */

static void init_arm(int n);
static Boolean move_angles_to(int player, float pivot, float arm, float elbow,
                float wrist, float wrist_rotate, float clutch, float drop);
static float vector_angle(float *a, float *b);
static void normalize(float *v);
static void thwok(void);


/****************************************************************************/
/**                                                                        **/
/**  This function initializes the two robot arms, setting up all objects  **/
/**  and constants necessary.                                              **/
/**                                                                        **/
/****************************************************************************/

void
init_arms(void)
{
    int i, j;

    /* Define some unit objects to be used to construct the arms. */

    coarse_disk = unit_disk(20, 1.0);
    medium_disk = unit_disk(28, 1.0);
    fine_disk = unit_disk(36, 1.0);
    chip_object = unit_disk(CHIP_SIDES, CHIP_THICKNESS);
    cube = unit_cube();

    /* Initialize the joint angles of the arms. */

    for (i=0; i<2; i++)
        for (j=0; j<ARM_PARTS; j++)
            angles[i][j] = default_angles[i][j];

    /* Calculate the distance from the wrist to the tips of the clutches. */

    clutch_size = sqrt((double)((scales[LOWER_FINGER][X] -
        (scales[UPPER_FINGER][X]/2.0)) * (scales[UPPER_FINGER][Y] -
        (scales[LOWER_FINGER][Y]/2.0))*2.0));

    /* Define where a chip should be relative to the wrist when it is */
    /* being carried.                                                 */

    chip_translate[X] = clutch_size;
    chip_translate[Y] = (CHIP_THICKNESS/2) - (scales[UPPER_FINGER][X]/2);
    chip_translate[Z] = 0.0;

    /* Initialize the display lists for each of the robot arms. */

    init_arm(0);
    init_arm(1);
}


/****************************************************************************/
/**                                                                        **/
/**  This function instructs the specified arm to have a chip in its       **/
/**  clutches.                                                             **/
/**                                                                        **/
/****************************************************************************/

void
move_chip(int arm)
{
    chip[arm] = chip_object;
}


/****************************************************************************/
/**                                                                        **/
/**  This function instructs the specified arm to not have a chip in its   **/
/**  clutches.                                                             **/
/**                                                                        **/
/****************************************************************************/

void
dont_move_chip(int arm)
{
    chip[arm] = -1;
}


/****************************************************************************/
/**                                                                        **/
/**  These functions dictate how fast the arms should move.  These         **/
/**  functions can be used to compensate for differences in machine        **/
/**  performance.  For example, on lower-end machines, fast_arm_speed()    **/
/**  would be a good setting, while on higher-end machines,                **/
/**  slow_arm_speed() would be a good setting.                             **/
/**                                                                        **/
/****************************************************************************/

void
slow_arm_speed(void)
{ degrees_per_frame = DEGREES_PER_SLOW_FRAME; }

void
medium_arm_speed(void)
{ degrees_per_frame = DEGREES_PER_MEDIUM_FRAME; }

void
fast_arm_speed(void)
{ degrees_per_frame = DEGREES_PER_FAST_FRAME; }


/****************************************************************************/
/**                                                                        **/
/**  This function instructs the specified arm to rotate its joints so     **/
/**  that it is in the correct orientation to pick up the chip at the      **/
/**  specified coordinates.  The orientation required is to have the       **/
/**  clutches parallel to the floor (and the chip) with the chip centred   **/
/**  in the open clutches.  The bottom of the clutches should line up with **/
/**  bottom of the chip so that the bottom of the clutches don't collide   **/
/**  with lower chips or the floor.                                        **/
/**                                                                        **/
/****************************************************************************/

Boolean
move_to_chip_pile(int arm, float *coords)
{
    float v[3], forward[3] = { 1.0, 0.0, 0.0 };
    float l, x, pivot_angle, elbow_angle, lower_arm_angle, wrist_angle;
    float w, a, d, q;

    x = (arm == 0)? -DISTANCE_FROM_CENTRE : DISTANCE_FROM_CENTRE;

    /* Figure out the base pivot angle.  Do this by finding the vector in */
    /* the XZ plane between the base of the robot arm and the chip, and   */
    /* taking the dot product between it and the "forward" vector of the  */
    /* base.                                                              */

    v[0] = coords[0] - x;
    v[1] = 0.0;
    v[2] = coords[2];
    q = l = (float) sqrt((double)(v[0]*v[0] + v[2]*v[2]));

    normalize(v);
    pivot_angle = vector_angle(v, forward);
    if (v[2] > 0.0) pivot_angle = -pivot_angle;

    /* Figure out the elbow angle. Do this by finding the 3-D vector    */
    /* between the base of the robot arm and the point where the wrist  */
    /* would have to be in order to be lined up with the chip, and then */
    /* performing a simple geometric calculation.                       */

    v[0] *= l - clutch_size;
    v[1] = coords[1]  + (scales[UPPER_FINGER][X]/2) -
                        (CHIP_THICKNESS/2) - scales[BASE][Y];
    v[2] *= l - clutch_size;
    l = (float) sqrt((double)(v[0]*v[0] + v[1]*v[1] + v[2]*v[2]));
    w = ((l*l - scales[UPPER_ARM][X]*scales[UPPER_ARM][X]
                        - scales[LOWER_ARM][X]*scales[LOWER_ARM][X]) /
                        -(2.0 * scales[UPPER_ARM][X]*scales[LOWER_ARM][X])) ;
    elbow_angle = acos((float)w) * (180.0 / PI);

    /* Correct the elbow angle to fit the needs of the display structure */

    elbow_angle = -(180.0 - elbow_angle);

    /* Figure out lower arm angle.  Do this by performing another funky */
    /* geometric calculation to find one angle and then finding the     */
    /* vector in the plane defined by the upper and lower arm between   */
    /* the base of the arm and the point where the wrist would have to  */
    /* in order to the lined up with the chip.  Then a dot product is   */
    /* taken with the "forward" direction to determine the lower arm    */
    /* angle.                                                           */

    w = ((scales[UPPER_ARM][X]*scales[UPPER_ARM][X] -
          scales[LOWER_ARM][X]*scales[LOWER_ARM][X] - l*l) /
          -(2.0 * scales[LOWER_ARM][X] * l));
    a = acosf(w) * (180.0 / PI);

    v[0] = q - clutch_size;
    v[1] = coords[1]  + (scales[UPPER_FINGER][X]/2) -
                        (CHIP_THICKNESS/2) - scales[BASE][Y];
    v[2] = 0.0;
    normalize(v);
    d = vector_angle(v, forward);
    if (v[1] > 0.0) d = -d;

    lower_arm_angle = a - d;

    /* Figure out the wrist angle by performing another funky geometric */
    /* calculation.                                                     */

    wrist_angle = asinf(scales[LOWER_ARM][X] *
                  (sinf((lower_arm_angle+d)*(PI/180.0)) / scales[UPPER_ARM][X]))
                  * (180.0/PI) + d;

    /* Move the arm to the orientation that was just calculated. */

    return move_angles_to(arm, pivot_angle,
                           lower_arm_angle,
                           elbow_angle,
                           wrist_angle,
                           0.0,
                           OPEN_ANGLE,
                           0.0);
}


/****************************************************************************/
/**                                                                        **/
/**  This function instructs the specified arm to pick up the chip that    **/
/**  should now be lying between the arm's clutches.  It does this simply  **/
/**  by closing the clutches on the chip.                                  **/
/**                                                                        **/
/****************************************************************************/

Boolean
pick_up_chip(int arm)
{
    chip_translate[X] = clutch_size;
    return move_angles_to(arm, angles[arm][BASE_PIVOT],
                           angles[arm][LOWER_ARM],
                           angles[arm][ELBOW],
                           angles[arm][WRIST],
                           0.0, 45.0, 0.0);
}


/****************************************************************************/
/**                                                                        **/
/**  This function instructs the specified arm to orient itself so that    **/
/**  it is in the position to drop the chip that it is presumably carrying **/
/**  into the specified column of the game board.  The orientation         **/
/**  required is to have the clutches parallel to the board (perpendicular **/
/**  to the ground) pointing downwards into the specified column at the    **/
/**  top of the game board.  It first moves to an intermediate position    **/
/**  to avoid colliding with the game board.                               **/
/**                                                                        **/
/****************************************************************************/

Boolean
move_to_column(int arm, int column)
{
    float pos[3], v[3], forward[3] = { 1.0, 0.0, 0.0 };
    float pivot_angle, elbow_angle, lower_arm_angle, wrist_angle,
          rotate_angle;
    float l, x, w, a, d, q;


    /* Move to an intermediate position to avoid colliding with the */
    /* game board.                                                  */

    if (!move_angles_to(arm, neutral_angles[arm][BASE_PIVOT],
                           neutral_angles[arm][LOWER_ARM],
                           neutral_angles[arm][ELBOW],
                           neutral_angles[arm][WRIST],
                           neutral_angles[arm][UPPER_ARM],
                           neutral_angles[arm][CLUTCHES], 0.0))
        return FALSE;

    /* Find out where we want the wrist to be. */

    coords_of_board_pos(column, 5, pos);
    pos[1] += DISTANCE_ABOVE_BOARD + clutch_size;

    x = DISTANCE_FROM_CENTRE + (scales[UPPER_FINGER][X]/2) - (CHIP_THICKNESS/2);
    if (arm == 0) x = -x;

    /* Figure out the base pivot angle.  Do this by finding the vector in */
    /* the XZ plane between the base of the robot arm and the desired     */
    /* position of the wrist, and taking the dot product between it and   */
    /* the "forward" vector of the base.                                  */

    v[0] = -x;
    v[1] = 0.0;
    v[2] = pos[2];
    q = l = (float) sqrt((double)(v[0]*v[0] + v[2]*v[2]));

    normalize(v);
    pivot_angle = vector_angle(v, forward);
    if (v[2] > 0.0) pivot_angle = -pivot_angle;

    /* Figure out the elbow angle. Do this by finding the 3-D vector    */
    /* between the base of the robot arm and the desired position of    */
    /* the wrist, and then performing a simple geometric calculation.   */

    v[0] *= l;
    v[1] = pos[1] - scales[BASE][Y];
    v[2] *= l;
    l = (float) sqrt((double)(v[0]*v[0] + v[1]*v[1] + v[2]*v[2]));
    w = ((l*l - scales[UPPER_ARM][X]*scales[UPPER_ARM][X]
                        - scales[LOWER_ARM][X]*scales[LOWER_ARM][X]) /
                        -(2.0 * scales[UPPER_ARM][X]*scales[LOWER_ARM][X])) ;
    elbow_angle = acos((float)w) * (180.0 / PI);

    /* Correct the elbow angle to fit the needs of the display structure */

    elbow_angle = -(180.0 - elbow_angle);

    /* Figure out lower arm angle.  Do this by performing another funky */
    /* geometric calculation to find one angle and then finding the     */
    /* vector in the plane defined by the upper and lower arm between   */
    /* the base of the arm and the desired position of the wrist.       */
    /* Then a dot product is taken with the "forward" direction to      */
    /* determine the lower arm angle.                                   */

    w = ((scales[UPPER_ARM][X]*scales[UPPER_ARM][X] -
          scales[LOWER_ARM][X]*scales[LOWER_ARM][X] - l*l) /
          -(2.0 * scales[LOWER_ARM][X] * l));
    a = acosf(w) * (180.0 / PI);

    v[0] = q;
    v[1] = pos[1] - scales[BASE][Y];
    v[2] = 0.0;
    normalize(v);
    d = vector_angle(v, forward);
    if (v[1] > 0.0) d = -d;

    lower_arm_angle = a - d;

    /* Figure out the wrist angle by performing another funky geometric */
    /* calculation.                                                     */

    wrist_angle = asinf(scales[LOWER_ARM][X] *
                  (sinf((lower_arm_angle+d)*(PI/180.0)) / scales[UPPER_ARM][X]))
                  * (180.0/PI) + d - 90.0;

    /* Figure out the wrist rotate angle.  This is a minor "twist" in the */
    /* clutches so it will line up exactly with the game board.           */

    v[0] = q;
    v[1] = pos[2];
    v[2] = 0.0;
    normalize(v);
    rotate_angle = vector_angle(v, forward);
    if ((arm == 0 && pos[2] > 0.0) || (arm == 1 && pos[2] < 0.0))
        rotate_angle = -rotate_angle;

    /* Move the arm to the orientation that was just calculated. */

    return move_angles_to(arm, pivot_angle,
                           lower_arm_angle,
                           elbow_angle,
                           wrist_angle,
                           rotate_angle,
                           45.0, 0.0);
}


/****************************************************************************/
/**                                                                        **/
/**  This function instructs the specified arm to drop the chip that is    **/
/**  presumably held in its clutches.  It does this by opening the         **/
/**  clutches and translating the chip forward by increasing intervals at  **/
/**  each frame (to give the appearance of acceleration) out to the        **/
/**  distance required to put the chip in the specified row.               **/
/**                                                                        **/
/****************************************************************************/

Boolean
drop_to_row(int arm, int row)
{
    return move_angles_to(arm, angles[arm][BASE_PIVOT],
                           angles[arm][LOWER_ARM],
                           angles[arm][ELBOW],
                           angles[arm][WRIST],
                           angles[arm][UPPER_ARM],
                           OPEN_ANGLE,
                           DISTANCE_ABOVE_BOARD + 5 - ((float) row));
}


/****************************************************************************/
/**                                                                        **/
/**  This function instructs the specified arm to return to its default    **/
/**  position.                                                             **/
/**                                                                        **/
/****************************************************************************/

Boolean
return_to_normal(int arm)
{
    return move_angles_to(arm,
        default_angles[arm][BASE_PIVOT],
        default_angles[arm][LOWER_ARM],
        default_angles[arm][ELBOW],
        default_angles[arm][WRIST],
        default_angles[arm][UPPER_ARM],
        default_angles[arm][CLUTCHES], 0.0);
}


/****************************************************************************/
/****************************************************************************/
/**                                                                        **/
/**  The following functions are local to this file.                       **/
/**                                                                        **/
/****************************************************************************/
/****************************************************************************/


/****************************************************************************/
/**                                                                        **/
/**  This function initializes the display list for the specified robot    **/
/**  arm.                                                                  **/
/**                                                                        **/
/****************************************************************************/

static void
init_arm(int n)
{
    float half_wrist;

    /* Upper finger */

    init_display_list(&upper_finger[n]);
    add_translate(&upper_finger[n],
                scales[LOWER_FINGER][X] - (scales[UPPER_FINGER][X] / 2.0), 
                -((scales[UPPER_FINGER][Y] / 2.0) -
                (scales[LOWER_FINGER][Y] / 2.0)), 0.0);
    add_scale(&upper_finger[n], scales[UPPER_FINGER]);
    add_object(&upper_finger[n], &cube);

    /* Lower finger */

    init_display_list(&lower_finger[n]);
    add_translate(&lower_finger[n], scales[LOWER_FINGER][X] / 2, 0.0, 0.0);
    add_scale(&lower_finger[n], scales[LOWER_FINGER]);
    add_object(&lower_finger[n], &cube);

    /* Finger */

    init_display_list(&finger[n]);
    add_rotate(&finger[n], NULL, NULL, &angles[n][CLUTCHES]);
    add_sublist(&finger[n], &lower_finger[n]);
    add_sublist(&finger[n], &upper_finger[n]);

    /* First finger */

    half_wrist = (1.0 + (sqrt(scales[LOWER_FINGER][Y] *
                 scales[LOWER_FINGER][Y]) / 2.0)) / 2.0;

    init_display_list(&first_finger[n]);
    add_translate(&first_finger[n], 0.0, half_wrist, 0.0);
    add_sublist(&first_finger[n], &finger[n]);
    

    /* Second finger */

    init_display_list(&second_finger[n]);
    add_translate(&second_finger[n], 0.0, -half_wrist, 0.0);
    add_rotate(&second_finger[n], &flip_angle, NULL, NULL);
    add_sublist(&second_finger[n], &finger[n]);

    /* Wrist */

    init_display_list(&wrist[n]);
    add_scale(&wrist[n], scales[WRIST]);
    add_object(&wrist[n], &coarse_disk);

    /* Current_chip */

    init_display_list(&current_chip[n]);
    add_material(&current_chip[n], chip_materials[n]);
    add_translatep(&current_chip[n], chip_translate);
    add_object(&current_chip[n], &chip[n]);

    /* Clutches */

    init_display_list(&clutches[n]);
    add_translate(&clutches[n], scales[UPPER_ARM][X], 0.0, 0.0);
    add_rotate(&clutches[n], NULL, NULL, &angles[n][WRIST]);
    add_rotate(&clutches[n], &angles[n][UPPER_ARM], NULL, NULL);
    add_sublist(&clutches[n], &current_chip[n]);
    add_rotate(&clutches[n], &square_angle, NULL, NULL);
    add_material(&clutches[n], arm_materials[n]);
    add_sublist(&clutches[n], &first_finger[n]);
    add_sublist(&clutches[n], &second_finger[n]);
    add_material(&clutches[n], METAL_MATERIAL);
    add_sublist(&clutches[n], &wrist[n]);

    /* Elbow */

    init_display_list(&elbow[n]);
    add_rotate(&elbow[n], &square_angle, NULL, NULL);
    add_scale(&elbow[n], scales[ELBOW]);
    add_material(&elbow[n], METAL_MATERIAL);
    add_object(&elbow[n], &coarse_disk);

    /* Upper Arm */

    init_display_list(&upper_arm[n]);
    add_translate(&upper_arm[n], scales[UPPER_ARM][X] / 2, 0.0, 0.0);
    add_scale(&upper_arm[n], scales[UPPER_ARM]);
    add_object(&upper_arm[n], &cube);

    /* Everything above lower arm */

    init_display_list(&all_above_lower[n]);
    add_translate(&all_above_lower[n], scales[LOWER_ARM][X], 0.0, 0.0);
    add_rotate(&all_above_lower[n], NULL, NULL, &angles[n][ELBOW]);
    add_sublist(&all_above_lower[n], &clutches[n]);
    add_sublist(&all_above_lower[n], &elbow[n]);
    add_material(&all_above_lower[n], arm_materials[n]);
    add_sublist(&all_above_lower[n], &upper_arm[n]);

    /* Lower arm */

    init_display_list(&lower_arm[n]);
    add_translate(&lower_arm[n], scales[LOWER_ARM][X] / 2, 0.0, 0.0);
    add_scale(&lower_arm[n], scales[LOWER_ARM]);
    add_object(&lower_arm[n], &cube);

    /* Everything above base pivot */

    init_display_list(&all_above_pivot[n]);
    add_rotate(&all_above_pivot[n], NULL, NULL, &angles[n][LOWER_ARM]);
    add_sublist(&all_above_pivot[n], &all_above_lower[n]);
    add_sublist(&all_above_pivot[n], &lower_arm[n]);

    /* Base pivot */

    init_display_list(&base_pivot[n]);
    add_rotate(&base_pivot[n], &square_angle, NULL, NULL);
    add_scale(&base_pivot[n], scales[BASE_PIVOT]);
    add_object(&base_pivot[n], &medium_disk);

    /* Everything but base */

    init_display_list(&all_but_base[n]);
    add_translate(&all_but_base[n], 0.0, scales[BASE][Y], 0.0);
    add_rotate(&all_but_base[n], NULL, &angles[n][BASE_PIVOT], NULL);
    add_sublist(&all_but_base[n], &all_above_pivot[n]);
    add_sublist(&all_but_base[n], &base_pivot[n]);

    /* Base */

    init_display_list(&base[n]);
    add_translate(&base[n], 0.0, scales[BASE][Y] / 2, 0.0);
    add_scale(&base[n], scales[BASE]);
    add_object(&base[n], &fine_disk);

    /* Robot Arm */

    init_display_list(&robot_arm[n]);
    add_material(&robot_arm[n], arm_materials[n]);
    add_sublist(&robot_arm[n], &all_but_base[n]);
    add_sublist(&robot_arm[n], &base[n]);
}


/****************************************************************************/
/**                                                                        **/
/**  This function returns the angle in degress between the two specified  **/
/**  3-D  vectors.  The returned angle will be in the range of 0 - 180.    **/
/**                                                                        **/
/****************************************************************************/

static float
vector_angle(float *a, float *b)
{
    return acosf(a[0] * b[0] + a[1] * b[1] + a[2] * b[2]) * (180.0/PI);
}


/****************************************************************************/
/**                                                                        **/
/**  This function normalizes the specified 3-D vector.                    **/
/**                                                                        **/
/****************************************************************************/

static void
normalize(float *v)
{
    float l;
    l = (float) sqrt((double)(v[0]*v[0] + v[1]*v[1] + v[2]*v[2]));
    v[0] /= l;
    v[1] /= l;
    v[2] /= l;
}


/****************************************************************************/
/**                                                                        **/
/**  This function animates the movement of the specified robot arm        **/
/**  (player) by moving the arm from its current orientation to the        **/
/**  orientation specified by the parameters.  If drop is non-zero, the    **/
/**  chip being held by the arm will be "dropped" by the specified         **/
/**  amount.                                                               **/
/**                                                                        **/
/****************************************************************************/

static Boolean
move_angles_to(int player, float pivot, float arm, float elbow, float wrist,
                        float wrist_rotate, float clutch, float drop)
{
    float pivot_diff, arm_diff, elbow_diff, wrist_diff, rotate_diff,
          clutch_diff, max_diff;
    float pivot_delta, arm_delta, elbow_delta, wrist_delta, rotate_delta,
          clutch_delta;
    float dropped = 0.0, drop_distance = 0.2;
    int i, frames;

    /* Determine the amount we have to rotate each angle. */

    pivot_diff = pivot - angles[player][BASE_PIVOT];
    arm_diff = arm - angles[player][LOWER_ARM];
    elbow_diff = elbow - angles[player][ELBOW];
    wrist_diff = wrist - angles[player][WRIST];
    rotate_diff = wrist_rotate - angles[player][UPPER_ARM];
    clutch_diff = clutch - angles[player][CLUTCHES];

    /* Make sure we take the shortest path. */

    if (pivot_diff > 180.0) pivot_diff -= 360.0;
    if (arm_diff > 180.0) arm_diff -= 360.0;
    if (elbow_diff > 180.0) elbow_diff -= 360.0;
    if (wrist_diff > 180.0) wrist_diff -= 360.0;
    if (rotate_diff > 180.0) rotate_diff -= 360.0;
    if (clutch_diff > 180.0) clutch_diff -= 360.0;

    if (pivot_diff < -180.0) pivot_diff += 360.0;
    if (arm_diff < -180.0) arm_diff += 360.0;
    if (elbow_diff < -180.0) elbow_diff += 360.0;
    if (wrist_diff < -180.0) wrist_diff += 360.0;
    if (rotate_diff < -180.0) rotate_diff += 360.0;
    if (clutch_diff < -180.0) clutch_diff += 360.0;

    /* Determine the maximum angle change that has to be made. */

    max_diff = 0.0;
    if (fabs(pivot_diff) > max_diff) max_diff = fabs(pivot_diff);
    if (fabs(arm_diff) > max_diff) max_diff = fabs(arm_diff);
    if (fabs(elbow_diff) > max_diff) max_diff = fabs(elbow_diff);
    if (fabs(wrist_diff) > max_diff) max_diff = fabs(wrist_diff);
    if (fabs(rotate_diff) > max_diff) max_diff = fabs(rotate_diff);
    if (fabs(clutch_diff) > max_diff) max_diff = fabs(clutch_diff);
    if (max_diff == 0.0) return;

    /* Determine the number of frames to be used for the animation. */

    frames = (int) (max_diff / degrees_per_frame) + 1;

    /* Determine the amount of change in each angle between each frame. */

    pivot_delta = pivot_diff / (float) frames;
    arm_delta = arm_diff / (float) frames;
    elbow_delta = elbow_diff / (float) frames;
    wrist_delta = wrist_diff / (float) frames;
    rotate_delta = rotate_diff / (float) frames;
    clutch_delta = clutch_diff / (float) frames;

    /* Draw each frame, incrementing each angle appropriately. */

    for (i=0; i<frames; i++) {
        angles[player][BASE_PIVOT] += pivot_delta;
        angles[player][LOWER_ARM] += arm_delta;
        angles[player][ELBOW] += elbow_delta;
        angles[player][WRIST] += wrist_delta;
        angles[player][UPPER_ARM] += rotate_delta;
        angles[player][CLUTCHES] += clutch_delta;

        /* Drop the chip by an appropriate amount if required. */

        if (dropped < drop) {
            dropped += drop_distance;
            drop_distance += 0.1;
            if (dropped >= drop) {
                 thwok();
                 dropped = drop;
            }
            chip_translate[X] = clutch_size + dropped;
        }

        /* Draw the world for this frame. */

        draw_world();

        /* Poll for an event.  If the event specified the ending of the */
        /* game, return the arm to its default position and abort the   */
        /* animation.                                                   */

        if (!get_event()) {
            dont_move_chip(player);
            return_to_normal(player);
            return FALSE;
        }
    }

    /* Make sure we're exactly where we want to be (to avoid accumulation */
    /* of tiny floating-point errors).                                    */

    angles[player][BASE_PIVOT] = pivot;
    angles[player][LOWER_ARM] = arm;
    angles[player][ELBOW] = elbow;
    angles[player][WRIST] = wrist;
    angles[player][UPPER_ARM] = wrist_rotate;
    angles[player][CLUTCHES] = clutch;

    /* If the chip hasn't dropped far enough yet, continue the drop. */

    while (dropped < drop) {
        dropped += drop_distance;
        drop_distance += 0.1;
        if (dropped >= drop) {
            thwok();
            dropped = drop;
        }
        chip_translate[X] = clutch_size + dropped;
        draw_world();
        if (!get_event()) {
            dont_move_chip(player);
            return_to_normal(player);
            return FALSE;
        }
    }

    return TRUE;
}


/****************************************************************************/
/**                                                                        **/
/**  This function creates a *thwok* sound.  It is system-dependent and    **/
/**  may have to be modified or removed if this program is ported to       **/
/**  another machine.                                                      **/
/**                                                                        **/
/****************************************************************************/

static void
thwok(void)
{
    static Boolean status = TRUE;
	if (status && system("playaiff tennis.aiff &") != 0)
		status = FALSE;
}

