/****************************************************************************/
/**                                                                        **/
/**  This file contains the functions necessary to define and draw a 7x6   **/
/**  game board.                                                           **/
/**                                                                        **/
/****************************************************************************/

#include <stdio.h>
#include <gl/gl.h>
#include "materials.h"
#include "unit_objects.h"
#include "display_list.h"
#include "chips.h"

#define BOTTOM_GAP      0.5

Display_list board, pick_slabs[7];

enum { X, Y, Z };

enum {
    VERTICAL_BAR,
    HORIZONTAL_BAR,
    BASE,
    BORDER,
    PICK_SLAB
};

static Object cube;
static Display_list horizontal_bar, vertical_bar, base, border,
                    horizontal_bars, vertical_bars, bases, borders;

/* Define the sizes of the various components of the game board. */

static float scales[5][3] = {
    /* X     Y    Z  */
    { 0.18, 6.0, 0.28 }, /* vertical bar */
    { 0.18, 0.22, 7.0 }, /* horizontal bar */
    { 2.4, 1.0, 1.4 },   /* base */
    { 0.7, 8.3, 0.4 },   /* border */
    { CHIP_THICKNESS, 7.0, 1.0 }  /* pick slab */
};


/****************************************************************************/
/**                                                                        **/
/**  This function initializes the display list for the both the game      **/
/**  board and the "pick slabs" used for picking purposes.                 **/
/**                                                                        **/
/****************************************************************************/

void
init_board(void)
{
    int i;

    cube = unit_cube();

    /* Vertical bar */

    init_display_list(&vertical_bar);
    add_translate(&vertical_bar, 0.0, (scales[VERTICAL_BAR][Y] / 2.0) +
                                scales[BASE][Y] + BOTTOM_GAP, 0.0);
    add_scale(&vertical_bar, scales[VERTICAL_BAR]);
    add_object(&vertical_bar, &cube);

    /* Horizontal bar */
    init_display_list(&horizontal_bar);
    add_scale(&horizontal_bar, scales[HORIZONTAL_BAR]);
    add_object(&horizontal_bar, &cube);

    /* Base */

    init_display_list(&base);
    add_scale(&base, scales[BASE]);
    add_object(&base, &cube);

    /* Border */

    init_display_list(&border);
    add_translate(&border, 0.0, scales[BORDER][Y] / 2.0, 0.0);
    add_scale(&border, scales[BORDER]);
    add_object(&border, &cube);

    /* Vertical bars */

    init_display_list(&vertical_bars);
    add_translate(&vertical_bars, 0.0, 0.0, -3.5);
    for (i=0; i<8; i++) {
        add_sublist(&vertical_bars, &vertical_bar);
        if (i < 7) add_translate(&vertical_bars, 0.0, 0.0, 1.0);
    }

    /* Horizontal bars */

    init_display_list(&horizontal_bars);
    add_translate(&horizontal_bars, 0.0, scales[BASE][Y] +
                (scales[HORIZONTAL_BAR][Y]/2) + BOTTOM_GAP, 0.0);
    for (i=0; i<7; i++) {
        add_sublist(&horizontal_bars, &horizontal_bar);
        if (i < 6) add_translate(&horizontal_bars, 0.0, 1.0, 0.0);
    }

    /* Borders */

    init_display_list(&borders);
    add_translate(&borders, 0.0, 0.0, -3.5 - scales[BORDER][Z]/2);
    add_sublist(&borders, &border);
    add_translate(&borders, 0.0, 0.0, 7.0 + scales[BORDER][Z]);
    add_sublist(&borders, &border);

    /* Bases */

    init_display_list(&bases);
    add_translate(&bases, 0.0, scales[BASE][Y] / 2, -3.5);
    add_sublist(&bases, &base);
    add_translate(&bases, 0.0, 0.0, 7.0);
    add_sublist(&bases, &base);

    /* The whole board */

    init_display_list(&board);
    add_material(&board, BOARD_BORDER_MATERIAL);
    add_sublist(&board, &borders);
    add_sublist(&board, &bases);
    add_material(&board, BOARD_MATERIAL);
    add_translate(&board, (scales[HORIZONTAL_BAR][X]+CHIP_THICKNESS) / 2,
                                                         0.0, 0.0);
    add_sublist(&board, &vertical_bars);
    add_sublist(&board, &horizontal_bars);
    add_translate(&board, -(scales[HORIZONTAL_BAR][X]+CHIP_THICKNESS),
                                                         0.0, 0.0);
    add_sublist(&board, &vertical_bars);
    add_sublist(&board, &horizontal_bars);

    /* The picking slabs */

    for (i=0; i<7; i++) {
        init_display_list(&pick_slabs[i]);
        add_translate(&pick_slabs[i], 0.0, (scales[PICK_SLAB][Y] / 2.0) +
                                scales[BASE][Y] + BOTTOM_GAP, (float) i - 3);
        add_scale(&pick_slabs[i], scales[PICK_SLAB]);
        add_object(&pick_slabs[i], &cube);
    }
}


/****************************************************************************/
/**                                                                        **/
/**  This function returns (in coords) the world coordinates of the exact  **/
/**  position of the space in the specified row and column of the game     **/
/**  board.                                                                **/
/**                                                                        **/
/****************************************************************************/

void
coords_of_board_pos(int column, int row, float *coords)
{
    if (column < 0 || column > 6 || row < 0 || row > 5) {
        fprintf(stderr, "coords_of_board_pos: invalid row/column\n");
        exit(1);
    }

    coords[0] = 0.0;
    coords[1] = (float) row + scales[BASE][Y] + BOTTOM_GAP +
                                (scales[HORIZONTAL_BAR][Y] / 2) + 0.5;
    coords[2] = (float) column - 3;
}

