/****************************************************************************/
/**                                                                        **/
/**  This file contains the functions necessary to define, draw and        **/
/**  manipulate two sets of 21 game chips.                                 **/
/**                                                                        **/
/****************************************************************************/

#include <stdlib.h>
#include <unistd.h>
#include <gl/gl.h>
#include "chips.h"
#include "unit_objects.h"
#include "materials.h"
#include "display_list.h"

Display_list chips;

enum { X, Y, Z };

/* Define the positions of the chip piles.  These are chosen wisely, so */
/* that the arms have no trouble reaching them.                         */

static float piles[2][2][3] = {

    /* Red */
    /*  X     Y     Z */
    {{ -5.0,  0.0, -7.0 },  /* first pile */
     {-10.0,  0.0, -9.0 }}, /* second pile */

    /* Blue */
    /*  X     Y     Z */
    {{  5.0,  0.0,  7.0 },  /* first pile */
     { 10.0,  0.0,  8.0 }}  /* second pile */
};

/* Some local variables. */

static Object chip;
static Object chip_objects[2][CHIPS_PER_SIDE];
static float orientations[2][CHIPS_PER_SIDE];
static float positions[2][CHIPS_PER_SIDE][3];
static Display_list individual_chip[2][CHIPS_PER_SIDE];
static int current_chip[2] = { 0, 0 };

static float purt(void);


/****************************************************************************/
/**                                                                        **/
/**  This function initializes the display list defining the positions     **/
/**  and orientations of all 42 game chips.                                **/
/**                                                                        **/
/****************************************************************************/

void
init_chips(void)
{
    int i, player;

    chip = unit_disk(CHIP_SIDES, CHIP_THICKNESS);

    /* Seed the random number generator with the pid of the current     */
    /* process to get a different staggering in the piles each time the */
    /* program is run.                                                  */

    srand(getpid());

    init_display_list(&chips);

    for (player=0; player<2; player++) {
        if (player == 0)
            add_material(&chips, RED_CHIP_MATERIAL);
        else
            add_material(&chips, BLUE_CHIP_MATERIAL);
        for (i=0; i<CHIPS_PER_SIDE; i++) {
            init_display_list(&individual_chip[player][i]);
            add_translatep(&individual_chip[player][i], positions[player][i]);
            add_rotate(&individual_chip[player][i],
                                 NULL, NULL, &orientations[player][i]);
            add_object(&individual_chip[player][i], &chip_objects[player][i]);
            add_sublist(&chips, &individual_chip[player][i]);
        }
    }

    init_chip_positions();
}


/****************************************************************************/
/**                                                                        **/
/**  This function initializes the position and orientation of each of     **/
/**  the 42 game chips into their respective piles (staggered randomly).   **/
/**                                                                        **/
/****************************************************************************/

void
init_chip_positions(void)
{
    int i, player, pile;
    float y;

    /* Create two piles for each player. */

    for (player=0; player<2; player++) {

        /* Place chips in the first pile. */

        y = (CHIP_THICKNESS * (CHIPS_PER_SIDE/2)) - (CHIP_THICKNESS / 2.0);
        for (i=0; i<CHIPS_PER_SIDE/2; i++) {
            chip_objects[player][i] = chip;
            positions[player][i][X] = piles[player][0][X] + purt();
            positions[player][i][Y] = y;
            positions[player][i][Z] = piles[player][0][Z] + purt();
            orientations[player][i] = 0.0;
            y -= CHIP_THICKNESS;
        }

        /* Place chips in the second pile. */

        y = (CHIP_THICKNESS * (CHIPS_PER_SIDE - (CHIPS_PER_SIDE/2))) -
                                                (CHIP_THICKNESS / 2.0);
        for (i=CHIPS_PER_SIDE/2; i<CHIPS_PER_SIDE; i++) {
            chip_objects[player][i] = chip;
            positions[player][i][X] = piles[player][1][X] + purt();
            positions[player][i][Y] = y;
            positions[player][i][Z] = piles[player][1][Z] + purt();
            orientations[player][i] = 0.0;
            y -= CHIP_THICKNESS;
        }
    }

    current_chip[0] = current_chip[1] = 0;
}


/****************************************************************************/
/**                                                                        **/
/**  This function indicates that the specified player is now interested   **/
/**  in the next chip in his/her pile(s).  It is assumed that there is     **/
/**  a next chip.                                                          **/
/**                                                                        **/
/****************************************************************************/

void
next_chip(int player)
{
    current_chip[player]++;
}


/****************************************************************************/
/**                                                                        **/
/**  This function instructs the current chip of the specified player to   **/
/**  not be drawn.  This is done when the robot arm has picked up the      **/
/**  chip, because it defines its own chip object to be translated with    **/
/**  the arm.                                                              **/
/**                                                                        **/
/****************************************************************************/

void
remove_current_chip(int player)
{
    chip_objects[player][current_chip[player]] = -1;
}


/****************************************************************************/
/**                                                                        **/
/**  This function instructs the current chip of the specified player to   **/
/**  be drawn.  This is done once the robot arm has finished dropping its  **/
/**  own version of the chip into the game board.                          **/
/**                                                                        **/
/****************************************************************************/

void
add_current_chip(int player)
{
    chip_objects[player][current_chip[player]] = chip;
}


/****************************************************************************/
/**                                                                        **/
/**  This function returns a pointer to the coordinates of the current     **/
/**  game chip of the specified player.                                    **/
/**                                                                        **/
/****************************************************************************/

float *
current_chip_pos(int player)
{
    return positions[player][current_chip[player]];
}


/****************************************************************************/
/**                                                                        **/
/**  This function specifies a new position for the current chip of the    **/
/**  specified player.  The chip is also rotated by 90 degrees so it can   **/
/**  fit inside the game board (its assumed destination).                  **/
/**                                                                        **/
/****************************************************************************/

void
new_chip_pos(int player, float *pos)
{
    positions[player][current_chip[player]][X] = pos[X];
    positions[player][current_chip[player]][Y] = pos[Y];
    positions[player][current_chip[player]][Z] = pos[Z];
    orientations[player][current_chip[player]] = 90.0;
}


/****************************************************************************/
/**                                                                        **/
/**  This local function returns a maximum purtubation to be applied to    **/
/**  X and Z coordinates of game chips as they are being placed in their   **/
/**  respective piles.  This creates a nice "staggering" effect.           **/
/**                                                                        **/
/****************************************************************************/

static float
purt(void)
{
    float p;

    p = ((float) rand() / (float) RAND_MAX) * MAX_PURTUBATION * 2;
    p = p - MAX_PURTUBATION;

    return p;
}
