/****************************************************************************/
/**                                                                        **/
/**  The following set of functions implements a display list.  The        **/
/**  display list allows the specification of a graphical scene in a       **/
/**  heirarchical fashion.                                                 **/
/**                                                                        **/
/****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <gl/gl.h>
#include "display_list.h"
#include "emalloc.h"

/****************************************************************************/
/**                                                                        **/
/**  This function must be called to initialize a Display_list before it   **/
/**  is used.                                                              **/
/**                                                                        **/
/****************************************************************************/

void
init_display_list(Display_list *display_list)
{
    display_list->first = NULL;
    display_list->last = NULL;
}


/****************************************************************************/
/**                                                                        **/
/**  This is a local function which helps the functions following it.  It  **/
/**  adds a new grNode element to the given Display_list, and returns a    **/
/**  pointer to the grNode so that it can be initialized with data.        **/
/**                                                                        **/
/****************************************************************************/

static grNode *
new_grNode(Display_list *display_list)
{
    grNode *node;

    node = (grNode *) emalloc(sizeof(grNode));

    if (display_list->last) {
        display_list->last->next = node;
        display_list->last = node;
    }
    else
        display_list->first = display_list->last = node;

    node->next = NULL;
    return node;
}


/****************************************************************************/
/**                                                                        **/
/**  This function will add a translate node to the Display_list.  The     **/
/**  action specified by this node is to translate the scene by x units    **/
/**  along the x-axis, y units along the y-axis, and z units along the     **/
/**  z-axis.                                                               **/
/**                                                                        **/
/****************************************************************************/

void
add_translate(Display_list *display_list, float x, float y, float z)
{
    grNode *node;

    node = new_grNode(display_list);
    node->type = TRANSLATE;
    node->info.translates[0] = x;
    node->info.translates[1] = y;
    node->info.translates[2] = z;
}


/****************************************************************************/
/**                                                                        **/
/**  This function will add a translatep node to the Display_list.  The    **/
/**  action specified by this node is to translate the scene by the data   **/
/**  specified in *p.                                                      **/
/**                                                                        **/
/****************************************************************************/

void
add_translatep(Display_list *display_list, float *p)
{
    grNode *node;

    node = new_grNode(display_list);
    node->type = TRANSLATEP;
    node->info.translatesp = p;
}


/****************************************************************************/
/**                                                                        **/
/**  This function will add a rotate node to the Display_list.  The action **/
/**  specified is to rotate the scene by *angle_x degrees around the       **/
/**  x-axis, then by *angle_y degrees around the y-axis, and finally by    **/
/**  *angle_z degrees around the z-axis.  If either of *angle_x, *angle_y, **/
/**  or *angle_z is NULL, no rotation occurs around that axis.             **/
/**                                                                        **/
/****************************************************************************/

void
add_rotate(Display_list *display_list, float *angle_x, float *angle_y,
                                                        float *angle_z)
{
    grNode *node;

    node = new_grNode(display_list);
    node->type = ROTATE;
    node->info.angles[0] = angle_x;
    node->info.angles[1] = angle_y;
    node->info.angles[2] = angle_z;
}


/****************************************************************************/
/**                                                                        **/
/**  This function will add a scale node to the Display_list.  The action  **/
/**  specified is to sale the scene by the scales provided in *scales.     **/
/**                                                                        **/
/****************************************************************************/

void
add_scale(Display_list *display_list, float *scales)
{
    grNode *node;

    node = new_grNode(display_list);
    node->type = SCALE;
    node->info.scales = scales;
}


/****************************************************************************/
/**                                                                        **/
/**  This function will add a material node to the Display_list.  The      **/
/**  action specified is to begin drawing objects with the surface         **/
/**  material specified by material_index.                                 **/
/**                                                                        **/
/****************************************************************************/

void
add_material(Display_list *display_list, int material_index)
{
    grNode *node;

    node = new_grNode(display_list);
    node->type = SURFACE_MATERIAL;
    node->info.material_index = material_index;
}


/****************************************************************************/
/**                                                                        **/
/**  This function will add an object node to the Display_list.  The       **/
/**  action specified is to draw the SGI Object pointed to by object.      **/
/**  This object will be subject to all transformations specified by the   **/
/**  nodes before it in the Display_list.                                  **/
/**                                                                        **/
/****************************************************************************/

void
add_object(Display_list *display_list, Object *object)
{
    grNode *node;

    node = new_grNode(display_list);
    node->type = DRAW_OBJECT;
    node->info.object = object;
}


/****************************************************************************/
/**                                                                        **/
/**  This function will add a sublist node to the Display_list.  The       **/
/**  action specified is to parse the Display_list pointed to by sublist.  **/
/**  The current transformation matrix is saved before the sublist is      **/
/**  parsed and restored afterwards so that any transformations specified  **/
/**  in the sublist will not affect future actions in the current          **/
/**  Display_list.                                                         **/
/**                                                                        **/
/****************************************************************************/

void
add_sublist(Display_list *display_list, Display_list *sublist)
{
    grNode *node;

    node = new_grNode(display_list);
    node->type = SUBLIST;
    node->info.sublist = sublist;
}


/****************************************************************************/
/**                                                                        **/
/**  This function will parse the provided Display_list, carrying out all  **/
/**  instructions specified by the nodes within.                           **/
/**                                                                        **/
/****************************************************************************/

void
do_display_list(Display_list *display_list)
{
    grNode *node;

    node = display_list->first;
    while (node) {
        switch (node->type) {
            case TRANSLATE:
                translate( node->info.translates[0],
                           node->info.translates[1],
                           node->info.translates[2]);
                break;
            case TRANSLATEP:
                translate( node->info.translatesp[0],
                           node->info.translatesp[1],
                           node->info.translatesp[2]);
                break;
            case ROTATE:
                if(node->info.angles[0] && *node->info.angles[0])
                    rot(*node->info.angles[0], 'x');
                if(node->info.angles[1] && *node->info.angles[1])
                    rot(*node->info.angles[1], 'y');
                if(node->info.angles[2] && *node->info.angles[2])
                    rot(*node->info.angles[2], 'z');
                break;
            case SCALE:
                scale(  node->info.scales[0],
                        node->info.scales[1],
                        node->info.scales[2]);
                break;
            case SURFACE_MATERIAL:
                lmbind(MATERIAL, node->info.material_index);
                break;
            case DRAW_OBJECT:
                if (*node->info.object >= 0)
                    callobj(*node->info.object);
                break;
            case SUBLIST:
                pushmatrix();
                do_display_list(node->info.sublist);
                popmatrix();
                break;
            default:
                fprintf(stderr, "do_display_list: illegal node type\n");
                exit(EXIT_FAILURE);
        }
        node = node->next;
    }
}

