/****************************************************************************/
/**                                                                        **/
/**  This file contains an interface to the trackball routines written by  **/
/**  Greg Hardy.  This interface was modified by Ramin Kamal.              **/
/**                                                                        **/
/****************************************************************************/

#include <gl.h>
#include <gl/device.h>
#include "tb_handler.h"
#include "trackball.h"

#define  DIR_NONE      0x00
#define  DIR_X         0x01
#define  DIR_Y         0x02
#define  DIR_Z         0x04

#define  SENS_PANX     30.0
#define  SENS_PANY     23.0
#define  SENS_ZOOM     35.0

Matrix mRotations    = {{1.0, 0.0, 0.0, 0.0},
                        {0.0, 1.0, 0.0, 0.0},
                        {0.0, 0.0, 1.0, 0.0},
                        {0.0, 0.0, 0.0, 1.0}},

       mTranslations = {{1.0, 0.0, 0.0, 0.0},
                        {0.0, 1.0, 0.0, 0.0},
                        {0.0, 0.0, 1.0, 0.0},
                        {0.0, 0.0, 0.0, 1.0}},

       mIdentity     = {{1.0, 0.0, 0.0, 0.0},
                        {0.0, 1.0, 0.0, 0.0},
                        {0.0, 0.0, 1.0, 0.0},
                        {0.0, 0.0, 0.0, 1.0}};

int nCurrentDir = DIR_NONE;

static void vCopyMatrix(Matrix mSource, Matrix mDestination);

void
ResetTrackBallMatrices()
{
        vCopyMatrix(mIdentity, mTranslations);
        vCopyMatrix(mIdentity, mRotations);
}

void
SetTrackBallLookat(     float vx, float vy, float vz, 
                        float px, float py, float pz, float twist)
{
        Matrix  tempMatrix;
        Matrix  lookMatrix;

        vCopyMatrix(mIdentity, mTranslations);
        vCopyMatrix(mIdentity, mRotations);

        getmatrix(tempMatrix);
        loadmatrix(mIdentity);
        lookat(vx, vy, vz, px, py, pz, twist);
        getmatrix(mRotations);

        mTranslations[3][0] = mRotations[3][0];
        mTranslations[3][1] = mRotations[3][1];
        mTranslations[3][2] = mRotations[3][2];

        mRotations[3][0] = 0.0;
        mRotations[3][1] = 0.0;
        mRotations[3][2] = 0.0;

        loadmatrix(tempMatrix);
}

static void 
vCopyMatrix(Matrix mSource, Matrix mDestination) 
{

  int i, j;

  for(i = 0; i < 4; i++)
    for(j = 0; j < 4; j++)
        mDestination[i][j] = mSource[i][j] ;
}

/*
 * Purpose   : Right multiplies matrix mMat1 by matrix mMat2 and stores
 *             the result in mMat1.
 */
static void 
vRightMultiply(Matrix mMat1, Matrix mMat2) {

   int    i, j;
   Matrix mMat3;

   for(i = 0; i < 4; i++)
      for(j = 0; j < 4; j++)
         mMat3[i][j] = mMat1[i][0] * mMat2[0][j] + mMat1[i][1] * mMat2[1][j] +
                       mMat1[i][2] * mMat2[2][j] + mMat1[i][3] * mMat2[3][j];
   for(i = 0; i < 4; i++)
      for(j = 0; j < 4; j++)
         mMat1[i][j] = mMat3[i][j] ;
}

static void 
vTranslate(float fTrans, char cAxis, Matrix mMat) 
{
   vCopyMatrix(mIdentity, mMat);
   switch(cAxis) 
   {
      case 'x':
         mMat[3][0] = fTrans;
         break;

      case 'y':
         mMat[3][1] = fTrans;
         break;

      case 'z':
         mMat[3][2] = fTrans;
         break;
   }
}

static void 
vToggleDir(int nDir) 
{
   nCurrentDir ^= nDir;
}

static void 
vPerformTransfo(float fOldX, float fNewX, float fOldY, float fNewY) 
{
   float  fRotVecX, fRotVecY, fRotVecZ;
   long         winwidth, winheight;
   Matrix mNewMat;
  
   getsize(&winwidth, &winheight);
   /*
    * Track ball rotations are being used.
    */
   if (nCurrentDir & DIR_Z) 
   {
      vGenerateRotVec(fNewX, fNewY, &fRotVecX, &fRotVecY, &fRotVecZ);
      vRotateScene(fRotVecX, fRotVecY, fRotVecZ, mNewMat);
      vRightMultiply(mRotations, mNewMat);
   }

   /*
    * Pan translations are being used.
    */
   if (nCurrentDir & DIR_X) 
   {
      vTranslate((fNewX - fOldX) / ((float)winwidth / SENS_PANX), 'x', 
                 mNewMat);
      vRightMultiply(mTranslations, mNewMat);
      vTranslate((fNewY - fOldY) / ((float)winheight / SENS_PANY), 'y', 
                 mNewMat);
      vRightMultiply(mTranslations, mNewMat);
   }
   
   /*
    * Zoom translations are being used.
    */
   if (nCurrentDir & DIR_Y) {
      vTranslate(-(fNewY - fOldY) / ((float)winheight / SENS_ZOOM), 'z', 
                 mNewMat);
      vRightMultiply(mTranslations, mNewMat);
   }
}


static void (*redrawer)() = NULL;

void
SetTrackBallRedrawer(void (*redrawFunc)())
{
        redrawer = redrawFunc;
}

void 
TrackBallMouseHandler()
{
        Device          device;
        short           startx, starty, currx, curry;
        float           fDummyX, fDummyY, fDummyZ;
        Matrix          mNewMat;
        int             left, middle, right;

        left = getvaluator(LEFTMOUSE);
        middle = getvaluator(MIDDLEMOUSE);
        right = getvaluator(RIGHTMOUSE);

        device = 0;
        if (right) device = RIGHTMOUSE;
        if (middle) device = MIDDLEMOUSE;
        if (left) device = LEFTMOUSE;

        nCurrentDir = DIR_NONE;

        while (left || middle || right)
        {
                if (device == LEFTMOUSE)
                        vToggleDir(DIR_X);
                if (device == MIDDLEMOUSE)
                        vToggleDir(DIR_Y);
                if (device == RIGHTMOUSE)
                        vToggleDir(DIR_Z);

                /*
                 * initialize the rotation vector.
                 */
                if (device != 0)
                {
                        float fDummyY2;

                        startx = getvaluator(MOUSEX);
                        starty = getvaluator(MOUSEY);
                        vGenerateRotVec((float)startx, (float)starty,
                                                &fDummyX, &fDummyY2, &fDummyZ);
                }

                if (nCurrentDir)
                {
                        currx = getvaluator(MOUSEX);
                        curry = getvaluator(MOUSEY);
                        vPerformTransfo((float)startx, (float)currx,
                                                (float)starty, (float)curry);
                        startx = currx;
                        starty = curry;
                        if(redrawer) redrawer();
                }

                device = 0;
                if (!left && getvaluator(LEFTMOUSE))
                        device = LEFTMOUSE;
                if (!middle && getvaluator(MIDDLEMOUSE))
                        device = MIDDLEMOUSE;
                if (!right && getvaluator(RIGHTMOUSE))
                        device = RIGHTMOUSE;
                left = getvaluator(LEFTMOUSE);
                middle = getvaluator(MIDDLEMOUSE);
                right = getvaluator(RIGHTMOUSE);
        }
}

void 
ApplyTrackBallTranslations() 
{
   multmatrix(mTranslations);
}

void 
ApplyTrackBallRotations() 
{
   multmatrix(mRotations);
}

void 
ApplyTrackBallTransformations() 
{
   multmatrix(mTranslations);
   multmatrix(mRotations);
}
