/*
 * Module : trackball.c
 *
 * Author : Michael Hardy 
 *
 * Date   : August 12, 1993
 *
 * Purpose: Contains routines to provide trackball-style 
 *          rotations.
 *
 */

#include <math.h>
#include <gl.h>

#include "trackball.h"

float fLastX, fLastY, fLastZ;

/*
 * Name      : void vRotateScene(float fVecX, float fVecY, float fVecZ, 
 *                               Matrix mNewMat)
 *
 * Parameters: float  fVecX   - The X component of the rotation vector.
 *             float  fVecY   - The Y component of the rotation vector.
 *             float  fVecZ   - The Z component of the rotation vector.
 *             Matrix mNewMat - The matrix to store the rotation in.
 *
 * Returns   : void
 *
 * Purpose   : Generates a matrix based on a rotation vector.
 */
void vRotateScene(float fVecX, float fVecY, float fVecZ, Matrix mNewMat) {

   float fScale = 3.0, fRadians, fInvLength, fNewVecX, fNewVecY, fNewVecZ,
         fSin0, fCos0, fVers0, fKXVers0, fKYVers0, fKZVers0, fKXSin0, 
         fKYSin0, fKZSin0;

   fVecX *= -1.0;
   fVecY *= -1.0;

   /* 
    * Find the length of the vector.
    */
    fRadians = sqrt(fVecX * fVecX + fVecY * fVecY + fVecZ * fVecZ);

   /*  
    * If the vector has zero length - return the identity matrix.
    */
   if (fRadians > -0.000001 && fRadians < 0.000001) {
      int i;
      int j;

      for (i = 0; i < 4; i ++) {
         for (j = 0; j < 4; j ++) {
            mNewMat[i][j] = (i == j ? 1.0 : 0.0);
         }
      }
      return;
   }

   /* 
    * Normalize the rotation vector.
    */
   fInvLength = 1 / fRadians;
   fNewVecX   = fVecX * fInvLength;
   fNewVecY   = fVecY * fInvLength;
   fNewVecZ   = fVecZ * fInvLength;
   fRadians  *= fScale;  

   /* 
    * Do all the smart stuff.
    */
   fSin0    = sin(fRadians);
   fCos0    = cos(fRadians);
   fVers0   = 1.0 - fCos0;
   fKXVers0 = fNewVecX * fVers0;
   fKYVers0 = fNewVecY * fVers0;
   fKZVers0 = fNewVecZ * fVers0;
   fKXSin0  = fNewVecX * fSin0;
   fKYSin0  = fNewVecY * fSin0;
   fKZSin0  = fNewVecZ * fSin0;

   mNewMat[0][0] = fNewVecX * fKXVers0 + fCos0;
   mNewMat[0][1] = fNewVecY * fKXVers0 - fKZSin0;
   mNewMat[0][2] = fNewVecZ * fKXVers0 + fKYSin0;
   mNewMat[0][3] = 0.0;
   mNewMat[1][0] = fNewVecX * fKYVers0 + fKZSin0;
   mNewMat[1][1] = fNewVecY * fKYVers0 + fCos0;
   mNewMat[1][2] = fNewVecZ * fKYVers0 - fKXSin0;
   mNewMat[1][3] = 0.0;
   mNewMat[2][0] = fNewVecX * fKZVers0 - fKYSin0;
   mNewMat[2][1] = fNewVecY * fKZVers0 + fKXSin0;
   mNewMat[2][2] = fNewVecZ * fKZVers0 + fCos0;
   mNewMat[2][3] = 0.0;
   mNewMat[3][0] = 0.0;
   mNewMat[3][1] = 0.0;
   mNewMat[3][2] = 0.0;
   mNewMat[3][3] = 1.0;
}

/*
 * Name      : void vGenerateRotVec(float fCurrentX, float fCurrentY, 
 *                                  float *fVecX, float *fVecY, float *fVecZ)
 *
 * Parameters: float fCurrentX - The current X location of the cursor.
 *             float fCurrentY - The current Y location of the cursor.
 *             float *fVecX    - The X component of the rotation vector.
 *             float *fVecY    - The Y component of the rotation vector.
 *             float *fVecZ    - The Z component of the rotation vector.
 *
 * Returns   : void
 *
 * Purpose   : Computes a rotation vector for the trackball
 *             interaction.
 */
void vGenerateRotVec(float fCurrentX, float fCurrentY, float *fVecX,
                     float *fVecY, float *fVecZ) 
{

        long nXOrigin, nYOrigin;
        long winwidth, winheight;
        long winposx, winposy;
        float fNewX, fNewY, fNewZ, fLength;

        getsize(&winwidth, &winheight);
        getorigin(&winposx, &winposy);
   /*
    * Compute center of window.
    */
   nXOrigin = winposx + winwidth / 2;
   nYOrigin = winposy + winheight / 2;

   /*
    * Calculate components.
    */
   fNewX    = (fCurrentX - (float)nXOrigin) * 2.0 / (float)winwidth;
   fNewY    = (fCurrentY - (float)nYOrigin) * 2.0 / (float)winheight;
   fNewZ    = (1.0 - fNewX * fNewX - fNewY * fNewY);

   /*
    * Moving around outside of window.
    */
   if (fNewZ < 0.0) {
      fLength = sqrt(1.0 - fNewZ);
      fNewX /= fLength;
      fNewY /= fLength;
      fNewZ  = 0.0;
   } else {
      fNewZ = sqrt(fNewZ);
   }

   /*
    * Generate rotation vector.
    */
   *fVecX = fLastY * fNewZ - fNewY * fLastZ;
   *fVecY = fLastZ * fNewX - fNewZ * fLastX;
   *fVecZ = -(fLastX * fNewY - fNewX * fLastY);

   /*
    * Update for next time.
    */
   fLastX = fNewX;
   fLastY = fNewY;
   fLastZ = fNewZ;
}
