/****************************************************************************/
/**                                                                        **/
/**  This file contains functions which define a number of unit-sized      **/
/**  objects to be used in this project.  The polygons defining these      **/
/**  objects are all specifiedin a counterclockwise manner so that         **/
/**  polygon backface removal techniques can be used.  Also, the normals   **/
/**  of each vertex in the objects are defined so that a lighting model    **/
/**  may be used.                                                          **/
/**                                                                        **/
/****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <gl/gl.h>

#define MAX_SIDES       200
#define PI              3.1415926536


/****************************************************************************/
/**                                                                        **/
/**  This function defines and returns a gl Object which resembles a disk  **/
/**  of unit diameter and of the specified thickness.  The "round" edge of **/
/**  the disk is simulated by a q-strip containing the specified number of **/
/**  polygons.                                                             **/
/**                                                                        **/
/****************************************************************************/

Object unit_disk(int sides, float thickness)
{
    register int i;
    float half_thickness;
    double delta, angle;
    double vertices[MAX_SIDES][2][3];
    float top_normal[3] = { 0.0, 1.0, 0.0 };
    float bottom_normal[3] = { 0.0, -1.0, 0.0 };
    float edge_normal[3];
    Object disk;

    if (sides > MAX_SIDES || sides < 3) {
        fprintf(stderr, "unit_disk: invalid number of sides\n");
        exit(EXIT_FAILURE);
    }

    half_thickness = thickness / 2.0;
    delta = (1.0 / (double) sides) * 2.0 * PI;

    /* Calculate the edge vertices required. */

    for (i=0, angle = 0.0; i<sides+1; i++, angle += delta) {
        vertices[i][0][0] = vertices[i][1][0] = sinf(angle) / 2.0;
        vertices[i][0][1] = half_thickness;
        vertices[i][1][1] = -half_thickness;
        vertices[i][0][2] = vertices[i][1][2] = cosf(angle) / 2.0;
    }

    makeobj(disk=genobj());

    /* Define the edge of the disk as a q-strip. */

    bgnqstrip();
        edge_normal[1] = 0.0;
        for (i=0; i<=sides; i++) {
            edge_normal[0] = vertices[i][0][0] * 2.0;
            edge_normal[2] = vertices[i][0][2] * 2.0;
            n3f(edge_normal);
            v3d(vertices[i][0]);
            v3d(vertices[i][1]);
        }
    endqstrip();

    /* Define the top of the disk as a polygon. */

    n3f(top_normal);
    bgnpolygon();
        for (i=0; i<sides; i++)
            v3d(vertices[i][0]);
    endpolygon();

    /* Define the bottom of the disk as a polygon. */

    n3f(bottom_normal);
    bgnpolygon();
        for (i=sides-1; i>=0; i--)
            v3d(vertices[i][1]);
    endpolygon();

    closeobj();

    return disk;
}


/****************************************************************************/
/**                                                                        **/
/**  This function defines and returns a gl Object which resembles a unit  **/
/**  cube, closed on all six sides.                                        **/
/**                                                                        **/
/****************************************************************************/

Object unit_cube(void)
{
    int i, j;
    Object cube;

    float surfaces[6][4][3] = {
        /*  All drawn counterclockwise with respect to the outward normal  */
        /*  of the surface.                                                */
        /* front */
        {{ .5,  .5,  .5}, {-.5,  .5,  .5}, {-.5, -.5,  .5}, { .5, -.5,  .5}},
        /* left */
        {{-.5,  .5,  .5}, {-.5,  .5, -.5}, {-.5, -.5, -.5}, {-.5, -.5,  .5}},
        /* back */
        {{-.5,  .5, -.5}, { .5,  .5, -.5}, { .5, -.5, -.5}, {-.5, -.5, -.5}},
        /* right */
        {{ .5,  .5, -.5}, { .5,  .5,  .5}, { .5, -.5,  .5}, { .5, -.5, -.5}},
        /* top */
        {{ .5,  .5, -.5}, {-.5,  .5, -.5}, {-.5,  .5,  .5}, { .5,  .5,  .5}},
        /* bottom */
        {{ .5, -.5,  .5}, {-.5, -.5,  .5}, {-.5, -.5, -.5}, { .5, -.5, -.5}},
    };

    float normals[6][3] = {
        { 0,  0,  1},   /* front */
        {-1,  0,  0},   /* left */
        { 0,  0, -1},   /* back */
        { 1,  0,  0},   /* right */
        { 0,  1,  0},   /* top */
        { 0, -1,  0}    /* bottom */
    };

    /* Define a "cube" object.  This cube is a unit cube centred at the */
    /* origin.  The polygons constructing the sides are defined in a    */
    /* counter-clockwise manner to allow the technique of polygon       */
    /* backface removal.                                                */

    makeobj(cube = genobj());
    for (i=0; i<6; i++) {
        n3f(normals[i]);
        bgnpolygon();
        for (j=0; j<4; j++)
            v3f(surfaces[i][j]);
        endpolygon();
    }
    closeobj();

    return cube;
}
