/****************************************************************************/
/**                                                                        **/
/**  This file contains the functions necessary to define the world as a   **/
/**  collection of its component parts.  Functions to initialize and draw  **/
/**  world are included.                                                   **/
/**                                                                        **/
/****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <gl/gl.h>
#include "tb_handler.h"
#include "unit_objects.h"
#include "arms.h"
#include "materials.h"
#include "chips.h"
#include "board.h"

#define MESSAGE_LENGTH  80

static Display_list arm0, arm1, world;
static char messages[2][MESSAGE_LENGTH] = { "", "" };
static Boolean draw_win = FALSE;
static int winx1, winy1, winx2, winy2;

enum {
    LIGHT_DEF = 1,
    LIGHT_MODEL,
    RED_CHIP_SURFACE,
    BLUE_CHIP_SURFACE
};

static float light_def[] = {
    LMNULL
};

static float light_model[] = {
    AMBIENT, 0.3, 0.3, 0.3,
    LMNULL
};

static void draw_announcement(void);


/****************************************************************************/
/**                                                                        **/
/**  This function initializes all of the display lists required for the   **/
/**  objects in the world.  A lighting model and a light source are also   **/
/**  defined.                                                              **/
/**                                                                        **/
/****************************************************************************/

void
init_world(void)
{
    lmdef(DEFLIGHT, LIGHT_DEF, 0, light_def);
    lmdef(DEFLMODEL, LIGHT_MODEL, 0, light_model);

    lmbind(LMODEL, LIGHT_MODEL);
    lmbind(LIGHT0, LIGHT_DEF);

    init_materials();
    init_arms();
    init_board();
    init_chips();

    init_display_list(&arm0);
    add_translate(&arm0, -8.0, 0.0, 0.0);
    add_sublist(&arm0, &robot_arm[0]);

    init_display_list(&arm1);
    add_translate(&arm1, 8.0, 0.0, 0.0);
    add_sublist(&arm1, &robot_arm[1]);

    init_display_list(&world);
    add_sublist(&world, &arm0);
    add_sublist(&world, &arm1);
    add_sublist(&world, &board);
    add_sublist(&world, &chips);
}


/****************************************************************************/
/**                                                                        **/
/**  This function draws the world by parsing the display list which       **/
/**  defines it.  A floor is also drawn.  If it has been specified by a    **/
/**  call to the draw_win_line() function that a winning set of game       **/
/**  pieces should be indicated, a white line is used to do this.          **/
/**                                                                        **/
/****************************************************************************/

void
draw_world(void)
{
    static float background[3] = {0.0, 0.0, 0.0};
    static float win_line[3] = {1.0, 1.0, 1.0};
    static float floor_normal[3] = {0.0, 1.0, 0.0};
    float win_1[3], win_2[3];
    static float vertices[4][3] = {
        { 18.0, 0.0, -16.0},
        {-18.0, 0.0, -16.0},
        {-18.0, 0.0,  16.0},
        { 18.0, 0.0,  16.0}
    };

    pushmatrix();
    ApplyTrackBallTransformations();
    c3f(background);
    clear();
    zclear();
    do_display_list(&world);
    lmbind(MATERIAL, FLOOR_MATERIAL);
    n3f(floor_normal);
    bgnpolygon();
        v3f(vertices[0]);
        v3f(vertices[1]);
        v3f(vertices[2]);
        v3f(vertices[3]);
    endpolygon();
    if (draw_win) {
        coords_of_board_pos(winx1, winy1, win_1);
        coords_of_board_pos(winx2, winy2, win_2);
        c3f(win_line);
        win_1[0] = win_2[0] = 0.3;
        bgnline();
            v3f(win_1);
            v3f(win_2);
        endline();
        win_1[0] = win_2[0] = -0.3;
        bgnline();
            v3f(win_1);
            v3f(win_2);
        endline();
    }
    draw_announcement();
    swapbuffers();
    popmatrix();
}


/****************************************************************************/
/**                                                                        **/
/**  This function loads an appropriate perspective matrix, insuring the   **/
/**  aspect ratio matches that of the current window.                      **/
/**                                                                        **/
/****************************************************************************/

void
my_perspective(void)
{
    long xsize, ysize;

    getsize(&xsize, &ysize);
    perspective(450, (float)xsize / (float)ysize, 0.20, 200.0);
}


/****************************************************************************/
/**                                                                        **/
/**  This function specifies a transformation matrix which places the eye  **/
/**  behind the red robot arm looking at the game board.                   **/
/**                                                                        **/
/****************************************************************************/

void
red_lookat(void)
{
    SetTrackBallLookat(-22.0, 4.6, -10.0, 0.0, 4.0, 0.0, 0.0);
}


/****************************************************************************/
/**                                                                        **/
/**  This function specifies a transformation matrix which places the eye  **/
/**  behind the blue robot arm looking at the game board.                  **/
/**                                                                        **/
/****************************************************************************/

void
blue_lookat(void)
{
    SetTrackBallLookat(22.0, 4.6, -10.0, 0.0, 4.0, 0.0, 0.0);
}


/****************************************************************************/
/**                                                                        **/
/**  This function specifies a transformation matrix which places the eye  **/
/**  above the game board looking down with the red robot arm to the left  **/
/**  and the blue robot arm to the right.                                  **/
/**                                                                        **/
/****************************************************************************/

void
top_lookat(void)
{
    SetTrackBallLookat(0.0, 33.0, 0.0, 0.0, 0.0, 0.0, 0.0);
}


/****************************************************************************/
/**                                                                        **/
/**  This function specifies a transformation matrix which places the eye  **/
/**  below the game board looking up with the red robot arm to the left    **/
/**  and the blue robot arm to the right.                                  **/
/**                                                                        **/
/****************************************************************************/

void
bottom_lookat(void)
{
    SetTrackBallLookat(0.0, -6.0, 0.0, 0.0, 0.0, 0.0, 0.0);
}


/****************************************************************************/
/**                                                                        **/
/**  This function allows the user to pick a column of the game board with **/
/**  the mouse.  It does this by entering gl's "pick" mode and drawing     **/
/**  "slabs" where the columns exist.  If one and only one slab is         **/
/**  selected, the corresponding column number is returned.  Otherwise     **/
/**  the computer beeps and a value of -1 is returned.                     **/
/**                                                                        **/
/****************************************************************************/

int
get_pick(void)
{
    short i, pickbuffer[7];
    long number_of_hits;
    int choice;
   
    /* Enter pick mode. */

    pick(pickbuffer, 7);
    my_perspective();

    /* Draw the "slabs". */

    zclear();
    for (i=0; i<7; i++) {
        loadname(i);
        pushmatrix();
        ApplyTrackBallTransformations();
        do_display_list(&pick_slabs[i]);
        popmatrix();
    }

    /* Leave pick mode. */

    number_of_hits = endpick(pickbuffer);
    my_perspective();

    /* If there were no hits, ring the bell and return -1. */

    if (number_of_hits < 1) {
        ringbell();
        return -1;
    }

    /* If more than one slab was selectedm ring the bell and return -1. */

    choice = pickbuffer[1];
    for (i=1; i < number_of_hits; i++)
        if (pickbuffer[i*2+1] != choice) {
            ringbell();
            return -1;
        }

    /* Return the column number selected. */

    return choice;
}


/****************************************************************************/
/**                                                                        **/
/**  This function specifies the string to draw on the specified message   **/
/**  line.  Note that this function doesn't actually draw the message.     **/
/**  The message will be drawn the next time draw_world() is called.       **/
/**                                                                        **/
/****************************************************************************/

void
announce(int msgnum, char *msg)
{
    if (msgnum < 0 || msgnum > 1) {
        fprintf(stderr, "announce: invalid message number\n");
        exit(EXIT_FAILURE);
    }

    strncpy(messages[msgnum], msg, MESSAGE_LENGTH);
}


/****************************************************************************/
/**                                                                        **/
/**  This function instructs that a thick white line should be drawn       **/
/**  between the specified rows and columns of the game board to indicate  **/
/**  a winning connection.  This line will be drawn on both sides of the   **/
/**  board.  Note that this function doesn't actually draw the line.  The  **/
/**  line will be drawn the next time draw_world() is called.              **/
/**                                                                        **/
/****************************************************************************/

void
draw_win_line(int x1, int y1, int x2, int y2)
{
    draw_win = TRUE;
    winx1 = x1;
    winy1 = y1;
    winx2 = x2;
    winy2 = y2;
}


/****************************************************************************/
/**                                                                        **/
/**  This function instructs that a thick white line should no longer be   **/
/**  drawn to indicate a winning connection.                               **/
/**                                                                        **/
/****************************************************************************/

void
hide_win_line(void)
{
    draw_win = FALSE;
}


/****************************************************************************/
/**                                                                        **/
/**  This function draws the announcements specified by prior announce()   **/
/**  function calls at the top-left corner of the window.  The message is  **/
/**  drawn in white.                                                       **/
/**                                                                        **/
/****************************************************************************/

static void
draw_announcement(void)
{
    static float announce_colour[3] = { 1.0, 1.0, 1.0 };
    static float identity[4][4] = {
        { 1, 0, 0, 0 },
        { 0, 1, 0, 0 },
        { 0, 0, 1, 0 },
        { 0, 0, 0, 1 }
    };

    pushmatrix();
    loadmatrix(identity);
    ortho2(0, 1, 0, 1);
    c3f(announce_colour);
    cmov2(0.01, 0.96);
    charstr(messages[0]);
    cmov2(0.01, 0.92);
    charstr(messages[1]);
    popmatrix();
    my_perspective();
}
