/******************************* BMTalkDialer.c *******************************/ 
/*
	Copyright  1993, Trustees of Dartmouth College
	
	Richard E. Brown
	
	Release date:	8 April 1993 

 This object implements a "modem dialer".
 
*/

#include <string.h>
#include <ctype.h>
#include <Serial.h>
#include <TBUtilities.h>
#include "histlog.h"
#include "MTalkUtilities.h"
#include "BMTalkDialer.h"

#pragma mark Enums, Consts, etc

enum errStrs {strNote=1,strCaution,strStop,strSys6,strMemFull,strNoop1,strDisrupt,
			strSysErr,strCantOpen,strCantSave,strNoDilr,strScriptErr,
			strBadInstr,strClicked,strNoop2,strInvSettings,strNoop3,
			strNoop4,strNoop5, strNoop6, strNoop7,strNoResp,strNoop8,
			strNoop9,strNoop10, andOthers };

enum BMTalkDialerCmds { sendCmd,receiveCmd,dialCmd,breakCmd,gotoCmd,
			ifCmd,countCmd,loopCmd,stopCmd,displayCmd,
			debugCmd,waitCmd,remCmd,initCmd,flushCmd,
			lastCmd };

#define numBMTalkDialerCmds lastCmd
#define PSLen(x) ((x) & 0x00FF)
#define numCmds lastCmd
#define kErrString 258
 
#define cmdList "SEND   RECEIVEDIAL   BREAK  GOTO   IF     COUNT  LOOP   STOP   DISPLAYDEBUG  WAIT   REM    INITSTRFLUSH  "

#define kDialDebug 1	/* turn off history of dialer commands */

/*****************************
 * I B M T A L K D I A L E R *
 *****************************
 
 Initialize the dialer.
 
*/

OSErr 	BMTalkDialer::IBMTalkDialer(BMTPrefsFile *thePrefsFile, BMTalkDLOG *theDisplay)
{
	OSErr	osstat;
	
	itsPrefs = thePrefsFile;
	itsDisplay = theDisplay;
	itsErrors = GetResource('STR#', kErrString);
	osstat = ResError();
	if (osstat) return(osstat);
	
	itsScript = itsPrefs->GetDialScript();	// get the script
	
}

/*****************
 * D I S P O S E *
 *****************
 
 Dispose of the Dialer's resources
 
*/

void	BMTalkDialer::Dispose(void)
{
	if (itsScript) ReleaseResource(itsScript);
	if (itsErrors) ReleaseResource(itsErrors);
	delete this;
}
	
/***********
 * D I A L *
 ***********
 
 Dial the desired script
 
*/

OSErr	BMTalkDialer::Dial(void)
{
 	OSErr	osstat;
 	short	modemNum;
 		
 	HLock(this);
 	HLock(itsScript);

	osstat = DDInit();
	if (osstat)		{		/* if trouble */
		DoAlert(Astop, 1, 3, 11, nil);	/* tell 'em */
	}
	else			{
		osstat = DDRun();		/* interpret the script */
	}

	HUnlock(this);
	HUnlock(itsScript);
	return(osstat);
}

/***************
 * H A N G U P *
 ***************
 
 Execute the script's hangup procedure. 
 
*/

OSErr	BMTalkDialer::Hangup(void)
{
	long	l;
	
	SendSlowString((char*)"\p+++");		// DEBUG! should send from script
	Delay (90, &l);
	SendSlowString((char*)"\pATH\r");
	
	return (noErr);
}
  
/*****************************************************************************
 DDInit - initialize the dialer
 *****************************************************************************/

 OSErr BMTalkDialer::DDInit(void)
 {
 	OSErr	osstat;
 	short	**hdl;
 	
 	hdl = (short**)itsScript;
 	pgmLen = **hdl;
 	pgmPtr = (Ptr) *hdl + 2;
 	pgmPCPtr = pgmPtr;
 	lastPC = 0;
 	dialPC = 1;
	
	return(noErr);
 }
 
 /*****************************************************************************
  DDRun - interpret a dialing script.  Return noErr if successful.
  *****************************************************************************/
 
 OSErr BMTalkDialer::DDRun (void)
 {
 	Boolean	Stop;
 	short	osstat;
 	char	errMsg[128];
 	char	num[10];
 	short	i;
// 	EventRecord theEvent;
 	
 	Stop = FALSE;					/* loop termination */
 	while(TRUE)	{
	 	if (AbortInQueue()) {			/* reason to stop? */
	 		FlushEvents(0xFFFF, 0);		// flush 'em all
			if (kDialDebug) histlog("Cmd-.",0,0);
	 		DoAlert(Anote, 1, 3, 14, NULL);/* then stop */
	 		itsDisplay->SetWatchCursor();
	 		Hangup();
	 		return(dialerAborted);
 		}
 		
 		osstat = DDFetch();
 		if (osstat)		{		/* get next instruction */
 			NumToString(dialPC, (StringPtr) &num);	/* if bad, get dialPC */
 			num[PSLen(num[0])+1] = 0;	/* make PZstring */
 			strcpy(&errMsg[1], "Tried to fetch instruction #");
 			strcat(&errMsg[1], &num[1]);
 			errMsg[0] = strlen(&errMsg[1]);	/* set the new Pstring length */
 			
 			if (kDialDebug) histlog("FetchErr",0,0);
 			DoAlert(Astop, 1, 3, 12, errMsg);
 			return(osstat);			/* ran off end of script */
 		}
 		
 		osstat = DDExecute(&Stop);		/* execute the instr */
 		if (osstat)		{		/* if trouble... */
 			if (osstat == dialerBadCmd) {	/* tell the user about bad commands */
			 	DoAlert(Astop, 1, 3, 13, statement);	
			}
 			return(osstat);			/* and exit no matter what */
 		}
 		else if (Stop)				/* script done? */
 			return(osstat);			/* return its result */
 	}
 	return(noErr);
 }
 
 /*****************************************************************************
  DDFetch - Fetch the next instruction 
  *****************************************************************************/

 short BMTalkDialer::DDFetch (void)
 {
 	register char	*p;
 	register short	i;
 	short	len, copyLen;
 	
 	p = pgmPCPtr;				/* point at "next" instruction */
 	if (dialPC > pgmLen) 			/* bad program - just quit */
 		return (dialerPastEOF);
 	else if (dialPC != (lastPC+1))	/* not next instruction in sequence */
 	{
 		p = pgmPtr;					 /* point back to the beginning */
		for (i = 1; i < dialPC; i++) /* and scan through the entire list */
		{
			p += PSLen(*p) + 1;		 /* add Pascal length to p to get next one */
		}
 	}
 	
 	/* p points at instruction to fetch */

	for (i = 0;i < sizeof(statement); i++)
		statement[i] = '';		/* fill with '' */
	len = PSLen(*p);			/* get the length */
	copyLen = (len < sizeof(statement)-2) ? len : sizeof(statement)-2;
	BlockMove(p, statement, copyLen+1);	/* copy the text */
	statement[copyLen+1] = 0;		/* tack on a nul */
	
	pgmPCPtr = p + len + 1;			/* update the pointers */		
	lastPC = dialPC;
	dialPC += 1;
	
 	return (noErr);
 }
  
 /*****************************************************************************
  DDExecute - Execute the instruction
  *****************************************************************************/
/*
 *	SEND 	<text to send to modem (to end of string)>
 *	RECEIVE <chars to wait for> <ticks to wait>
 *	DIAL	
 *	BREAK
 *	GOTO	<instruction to go to>
 *	IF	<string to search for in receive buffer> <instruction to goto>
 *	COUNT	<number of times to loop>
 *	LOOP	
 *	STOP	<status to return>
 *	DISPLAY <text to display (to end of string)>
 *	DEBUG	<on/off>
 *	WAIT	<ticks to wait>
 *	REM
 *	INITSTR
 *	FLUSH  
 */

short tokCnt[] = {1, 2, 0, 0, 1, 2, 1,     0, 1, 1, 1, 2, 0, 0, 0 };

short BMTalkDialer::DDExecute (Boolean * stop)
{
 	short 	nextChar;
 	short	retval;
 	short	keyWord;
 	
 	nextChar = 1;				/* starting at first char */
 	keyWord = MatchKey(&nextChar);		/* is it a keyword? */
 	if (keyWord < 0)
 	{	
 		return(dialerBadCmd);
 	}
 	
 	switch(keyWord)				/* strip the comments */	
 	{
 	case sendCmd: 
 	case displayCmd: 
 				break;			/* except in literal commands */
 	default: 	StripComment(&nextChar);
 	}
 		
 	ParseTokens(tokCnt[keyWord], nextChar);	/* grab the token(s) */
 	
	if (kDialDebug) histlog(&statement[1], *(long*)token[0], *(long*)token[1]);
 	
 	switch(keyWord)		{
	case sendCmd:
		retval = DoSend(stop); break;
	case receiveCmd:
		retval = DoReceive(stop); break;
	case dialCmd:
		retval = DoDial(stop); break;
	case breakCmd:
		retval = DoBreak(stop); break;
	case gotoCmd:
		retval = DoGoto(stop); break;
	case ifCmd:
		retval = DoIf(stop); break;
	case countCmd:
		retval = DoCount(stop); break;
	case loopCmd:
		retval = DoLoop(stop); break;
	case stopCmd:
		retval = DoStop(stop); break;
	case displayCmd:
		retval = DoDisplay(stop); break;
	case debugCmd:
		retval = DoDebug(stop); break;
	case waitCmd:
		retval = DoWait(stop); break;
	case remCmd:
		retval = DoRem(stop); break;
	case initCmd:
		retval = DoInitStr(stop); break;
	case flushCmd:
		retval = DoFlush(stop); break;
 	}
 	return(retval);
 }

/*****************************************************************************
 DoSend - send the first token to the modem
 *****************************************************************************/
 
short 	BMTalkDialer::DoSend(Boolean * stop)
{
	long	*lp;
	
	lp = (long*)token[0];
	if (kDialDebug) histlog("Sending",*lp,TickCount());
	ClearModemInput();			/* flush modem input */
	SendSlowString((char *) token[0]);	/* send the token (Pstring) */
	return(noErr);
}
/*****************************************************************************
 DoReceive - wait 'til the string arrives or time out after n ticks
 *****************************************************************************/
 
short 	BMTalkDialer::DoReceive(Boolean * stop)
{
	long	timeout;
	long	t,t1;
	register short	i;
	short	osstat;
	char	c;
	long	numChars;
	long	l, l1, *lp;
	char	*p;
	
	StripBlanks(token[1]);		/* remove blanks from the time */
	makelc(&token[0][1]);		/* make the pattern lc */
	StringToNum((StringPtr) &token[1], &timeout);	/* get the time value */
	t = TickCount();
	while (TRUE)   {	
		t1 = TickCount();
		if (AbortInQueue()) 	{	/* Cmd-.? */
			if (kDialDebug) histlog("cmd-.",timeout,TickCount());
			break;
		}
		if ((t1-t) > timeout)	{ /* if we timed out */
			BlockMove(modemInput, matchString, PSLen(*modemInput)+2);
			if (kDialDebug) histlog("Timeout",t1-t,timeout);	// entire string is matchString
			break;
		}
		while (TRUE)		{	/* suck up characters from modem */
			t1 = TickCount();

			osstat = GetChar(&c);
			if (osstat == eofErr) 	  /* no data now */
				break;
			i = PSLen(*modemInput)+1; /* next free byte */
			if (i < sizeof(modemInput)-1) {	/* leave last byte NUL */
//				if (kDialDebug) histlog("got char",(i<<16)+c,t1);
				modemInput[i] = tolower(c);
				modemInput[0] = i;	/* keep length up to date */
				modemInput[i+1] = 0;	/* tack NUL at end */
			}
		}
		
		/* sucked up all the characters for now; check for a match */
		
		p = strstr(&modemInput[1], &token[0][1]);		
		if (p)  {	 			/* if it matches...*/
			BlockMove(modemInput, matchString, PSLen(*modemInput)+2);
			l = p - (char*)&modemInput; 	/* length of matching stuff */
			l += PSLen(token[0][0]);
			matchString[0] = l-1;		/* Pascal length byte */
			makePZString(matchString);	/* make it PZString */
			
			l1 = PSLen(*modemInput) - l + 1;	
			BlockMove(&modemInput[l], &modemInput[1],l1);
			modemInput[0] = l1;
			for (i = l1+1;i < sizeof(modemInput); i++)
				modemInput[i] = 0;	/* clear the remainder */

			lp = (long*)token[0];
			if (kDialDebug) histlog("RCVmatch",*lp, TickCount());
			break;				/* break out of timed loop */
		}
	}
	return(noErr);
}

/*****************************************************************************
 DoDial - Dial the phone number
 *****************************************************************************/
 
short 	BMTalkDialer::DoDial(Boolean * stop)
{
	char	dialStr[256];
	char	needsCC[20];
	Boolean myStop;

	ClearModemInput();
	itsPrefs->GetAPref(sDoCCard, needsCC, sizeof(needsCC));
	itsPrefs->GetDialString(dialStr);	
	SendSlowString(dialStr);	/* send their phone number (Pstring) */
	if (kDialDebug) histlog("dialing",0,TickCount());
	
	
// if calling card is necessary, wait for "OK", then send CC number

	if (needsCC[0])	{		// if there's a calling card number
		PZstrcpy(token[0], (char*)"\pOK", 10);	// wait up to 30 sec
		PZstrcpy(token[1], (char*)"\p1800", 10);//   for "OK" response
		DoReceive(&myStop);		// RECEIVE "OK" 1800
		itsPrefs->GetCCString(dialStr);
		SendSlowString(dialStr); // and then send the CC number
	}
	return(noErr);
}

/*****************************************************************************
 DoBreak - send a break to the modem
 *****************************************************************************/
 
short 	BMTalkDialer::DoBreak(Boolean * stop)
{
	OSErr	osstat;
	short	i;
	
	ClearModemInput();
//	i = -1;					/* break is a char of "-1" */
//	osstat = AA_Control(putChar, &i);	/* send it */
	return(osstat);
}

/*****************************************************************************
 DoGoto - go to a specific instruction
 *****************************************************************************/
 
short 	BMTalkDialer::DoGoto(Boolean * stop)
{
	long	*lp;
	
	if (kDialDebug) histlog("goto ",*(long*)&token[0],0);
	UpdatePC(token[0]);
	return(noErr);
}

/*****************************************************************************
 DoIf - if modemInput contains the string, goto the indicated instruction
 *****************************************************************************/
 
short 	BMTalkDialer::DoIf(Boolean * stop)
{
	short	i;
	long	*lp;
	
	i = 0;
	makelc(&token[0][1]);		/* make the test string lc */
	
	if (strstr(&matchString[1], &token[0][1]))
	{
		UpdatePC(token[1]);
		lp = (long*)token[0];
 		if (kDialDebug) histlog("IFmatch",*lp, dialPC);
	}
	return(noErr);
}

/*****************************************************************************
 DoCount - loop some number of times
 *****************************************************************************/
 
short 	BMTalkDialer::DoCount(Boolean * stop)
{
	long	l;
	
	StripBlanks(token[0]);				/* remove blanks from the token */
	StringToNum((StringPtr) &token[0], &l);			/* get the new counter value */
	counter = l;						/* set the counter */
	loopPC = dialPC;					/* and set the loopPC */
	return(noErr);
}

/*****************************************************************************
 DoLoop - go back to the top of the loop
 *****************************************************************************/
 
short 	BMTalkDialer::DoLoop(Boolean * stop)
{
	
	if (counter == 0)	{ 	/* nothing - just fall thru this statement */ }
	else 					/* gotta loop */
	{
		counter--;			/* decr the counter */
		dialPC = loopPC;	/* and jump back to the top */
	}
		
	return(noErr);
}

/*****************************************************************************
 DoDisplay - display the string in the status box on the screen
 *****************************************************************************/
 
short 	BMTalkDialer::DoDisplay(Boolean * stop)
{
	if (token[0][0] == 0)					/* if no arguments */
		itsDisplay->AddToDisplay((char *) &modemInput);	/* 	display what the modem said */
	else 
		itsDisplay->AddToDisplay((char *) &token[0]);	/* otherwise show first token */
	return(noErr);
}

/*****************************************************************************
 DoStop - stop this thing
 *****************************************************************************/
 
short 	BMTalkDialer::DoStop(Boolean * stop)
{
	long	l;
	short	retval;
	
	*stop = 1;				/* remember to stop */
	StripBlanks(token[0]);			/* remove blanks from the token */
	StringToNum((StringPtr) token[0], &l);	/* get return value */
	retval = l;
	return(retval);
}

/*****************************************************************************
 DoDebug - turn on/off debugging
 *****************************************************************************/
 
short 	BMTalkDialer::DoDebug(Boolean * stop)
{
/*	DDebugStr("\pExecuted Debug instruction");*/
	BlockMove(&token[0], modemInput, PSLen(token[0][0])); /* and crock up modemInput */
	modemInput[PSLen(token[0][0])] = 0; 			/* add trailing null */
	makelc(&modemInput[1]);
	return(noErr);
}

/*****************************************************************************
 DoWait - wait some number of ticks
 *****************************************************************************/
 
short 	BMTalkDialer::DoWait(Boolean * stop)
{
	long	t;
	long	timeout;
	EventRecord theEvent;
	
	StripBlanks(token[0]);				/* remove blanks from the token */
	StringToNum((StringPtr) &token[0], &timeout);
	t = TickCount();
	while ((TickCount()-t) < timeout)	/* if we haven't had enough ticks */
	{
		if(MyWaitNextEvent(1<<nullEvent,&theEvent,1,NULL)) { }
		if (Button()) break;			/* button says we should pretend we're done */
	}
	return(noErr);
}

/*****************************************************************************
 DoRem - don't do anything...
 *****************************************************************************/
 
short 	BMTalkDialer::DoRem(Boolean * stop)

{
	return(noErr);						/* speedy, eh? */
}

/*****************************************************************************
 DoInitStr - send the modem's init string, followed by a CR
 *****************************************************************************/
 
short 	BMTalkDialer::DoInitStr(Boolean * stop)
{
	char	str[kMTMaxPref];

	itsPrefs->GetModemString(str);	/* get the modem's INIT string */
	ClearModemInput();
	SendSlowString((char*) str);	/* send their init. string (Pstring) */
	if (kDialDebug) histlog("Initstr",0,TickCount());
	return(noErr);
}

/*****************************************************************************
 DoFlush - Clear the modem's input buffer
 *****************************************************************************/
 
short 	BMTalkDialer::DoFlush(Boolean * stop)
{
	ClearModemInput();
	return(noErr);
}

/*****************************************************************************
 MatchKey - return the index of the keyword, or -1
 *****************************************************************************/

short	BMTalkDialer::MatchKey(short * nextChar)
{
	register short	theCmd;
	register short 	i;
	register char	*str;
	
	for (i = 1; statement[i] != 0; i++)
	{
		if (statement[i] == ' ') break;		/* scan 'til space */
		statement[i] = toupper(statement[i]);	/* change to be UPPER CASE */
	}
	
	/* i is first non-token character */
	
	*nextChar = i+1;				/* remember this */
	i--;										/* and back up */
	
	str = cmdList;
 	for (theCmd = 0; theCmd < numCmds; theCmd++) 	/* search out the command */
 	{
	 	if (strncmp(&statement[1],str,i) == 0) 	/* if they're the same */
	 		break;								/* 	then break out */
	 	str += 7;
 	}

	if (theCmd == numCmds) theCmd = -1;		/* doesn't match a command */
	return(theCmd);					/* say what it is */
}

/*****************************************************************************
 StripComment - strip off comments from the line (from the rightmost "!")
 *****************************************************************************/

void	BMTalkDialer::StripComment(short * nextChar)
{
	register short	i;
	register short 	j;
	
	j = *nextChar;
	for (i = PSLen(statement[0]); i > j; i--)	/* scan from the right */
	{	
		if (statement[i] == '!') break;		/* 'til a '!' */
	}
	
	if (i != j)
	{
		statement[i] = 0;			/* add trailing nul */
		statement[0] = i-1;			/* don't strip if no "!" */
	}
	
}

/*****************************************************************************
 ParseTokens - grab zero, one or two tokens from the statement
 *****************************************************************************/

void	BMTalkDialer::ParseTokens(short numArgs, short nextChar)

{
	register char	*s,*d;		/* source and destination pointers */
	register short	i,j;
	
	if (numArgs > 0)		/* scan off at least one argument */
	{
		s = &statement[nextChar];
		d = &token[0][1];
		i = nextChar;
	 
		while (i <= PSLen(statement[0])) /* scan first arg: entire line if only one */
		{
			if ((numArgs != 1) && (*s == ' ')) break;
			*d++ = *s++;
			i++;
		}
		*d = 0;			/* tack null on end of string */
		token[0][0] = (i-nextChar);	/* and set up PString length */
		
	}
	if (numArgs == 2)
	{	
		/* drop intervening blanks */
		
		while(*s == ' ')   { s++; i++; }
		
		j = i;
		d = &token[1][1];
		while ((*s != ' ') && (i <= PSLen(statement[0])))	
		{
			*d++ = *s++;
			i++;
		}
		*d = 0;
		token[1][0] = (i-j);
	}
	RemoveDollars(token[0]);
	RemoveDollars(token[1]);
}
  
/*****************************************************************************
 atoh - ascii char to hex nibble
 *****************************************************************************/

short	BMTalkDialer::atoh(char * theChar)

{
	register char	c;
	
	c = *theChar;
	if (c <= '9') 
		 return (c - '0');
	if (c <= 'F') 
		 return (c - 'A' + 10);
	else return (c - 'a' + 10);
}

/*****************************************************************************
 RemoveDollars - take out $xx stuff
 *****************************************************************************/

void	BMTalkDialer::RemoveDollars(char * theTok)

{
	char	*s, *d;		
	short	len, i, j;
	
	s = theTok;
	s++;					/* point at first char */
	d = s;
	j = 0;
	while (*s != 0)			/* loop thru the string */
	{	
		if (*s != '$')		/* if not the escape character */
		{
			*d++ = *s++;	/* copy the char & bump pointers */
		}		
		else				/* it's a '$' - consume two more chars */
		{
			s++;			/* move to next char */
			i = atoh(s);	/* get its value */
			s++;
			i = i*16 + atoh(s); /* and the next */
			s++;
			*d++ = i;		/* and save it in the buffer */
		}
		j++;				/* bump the (new) length */
	}
	*d = 0;					/* tack on the nul */
	*theTok = j;			/* & update the new length (PString) */
}

/*****************************************************************************
 makelc - make the str lower case, and return the last char we fixed.
 *****************************************************************************/
 
void 	BMTalkDialer::makelc(char * theStr)
{
	char	*p;
	
	for (p = theStr; *p != 0; p++)
	{
		*p = tolower(*p);
	}
}

/*****************************************************************************
 StripBlanks - strip leading & trailing blanks from a PZstring
 *****************************************************************************/
 
void 	BMTalkDialer::StripBlanks(char * theStr)
{
	char	*s, *d;
	short	i;
	
	s = theStr;
	s++; 								/* point at first real char */
	d = s;
	while ((*s != 0) && (*s == ' ')) 	/* scan to first non-blank */
		s++; 
		
	i = 0;								/* i is the string length */
	while ((*s != 0) && (*s != ' '))	/* scan to next space */
	{
		*d++ = *s++;
		i++;
	}
	*d = 0;								/* tack on trailing nul */
	*theStr = i;						/* and update the length byte */
}

/*****************************************************************************
 ClearModemInput - flush modem's PZstring, eat up queued input characters
 *****************************************************************************/
 
void 	BMTalkDialer::ClearModemInput(void)
{
	register short i;
	char	c;
	long	numChars;
	OSErr	osstat;
	long	t;
	
	for (i=0;i<sizeof(modemInput); i++) 	/* clear existing string */
		modemInput[i] = 0;
	t = TickCount();			/* clear out pending data */
	while (TickCount()-t < 6)	{	/* wait six ticks (1/10 second) */
		osstat = GetChar(&c);		/* try for another char */
		if (osstat != eofErr)  {	/* Got one */
			t = TickCount();	/* and reset the timer */
			if (kDialDebug) histlog("skipping",c,t);
		}
	}
}

/*****************************************************************************
 UpdatePC - fix up PC for goto, if, or receive command 
 *****************************************************************************/
 
void 	BMTalkDialer::UpdatePC(char * theAddr)
{
	long	l;

	StripBlanks(theAddr);
	StringToNum((StringPtr) theAddr, &l);
	switch(*(theAddr+1))
	{
	case '+': case '-':			/* if it's relative, add to PC */
		dialPC = dialPC + l - 1;	/*   (dialPC is already past this instr ) */
		break;
	default:
		dialPC = l;			/* otherwise, replace it */
		break;
	}
}

 /*****************************************************************************
  DoAlert - display the right text in the user interface
  *****************************************************************************/
 
short 	BMTalkDialer::DoAlert(short alertType, short numButtons, 
 				short strNum1, short strNum2, char *extraStr)

 {
 	char 	str1[258];
 	char	str2[258];
 	
	GetIndString(str1,kErrString,strNum1); 	// Get first string
	makePZString(str1);			// make PZString
	PZstrcat(str1, (char *)"\p ", sizeof(str1));	// append a space
	GetIndString(str2,kErrString,strNum2); 	// Get second string
	makePZString(str2);			// make PZString
	PZstrcat(str1, str2,sizeof(str1));	// Tack on second string
	
	if (extraStr)	{			//
		PZstrcat(str1, (char*)"\p ", sizeof(str1));
	 	PZstrcat(str1,str2,sizeof(str1));
	}
	
	itsDisplay->AddToDisplay(str1);		// show it in dialog... 	
 }