/*************************** MTalkWrite.c *************************************/ 
/*
	Copyright  1993, Trustees of Dartmouth College
	
	Richard E. Brown
	
	Release date:	8 April 1993 

 This file implements the LAPWrite/LAPEnq functions.
 
*/
#include 	<Serial.h>
#include 	<Retrace.h>
#include 	<AppleTalk.h>
#include	"atlk.h"
#include	"ModemTalk.h"
#include 	"MTalkUtilities.h"
#include 	"mySetupA4.h"			// local copy
#include	"LAPMgrGlue.h"
#include 	"histlog.h"

#pragma mark Internal functions

void	BuildAAwds(WDSElement* aWDS, Ptr dataBuf, short lapT, 
				short theLen, Ptr theData);
long	Frame(WDSElement* aWDS, char * theBuf);
OSErr	SendOutBuf(ProcPtr theWriteCompl);
char*	Escape(unsigned char *p, char * d);
void	RewriteHeader(WDSElement *wds1, WDSElement *wds2, char *buf);


/***************
 * L A P W R T *
 ***************
  
 This code will be called when data packet is to be written to an AppleTalk
 address. The parameters, passed in an array of registers, are:
 
 	rA0 = return address after the packet has been written out
 	rA1 -> WDS which describes the data to be written
 	rA2 ->.MPP variables
 	rD2 = AppleTalk destination address specifier (byte, for ModemTalk)
 	rD3 = port to write through
 
 The LAPWrt code returns the following values:
 	 0 = noErr, and the operation completed
 	<0 = error occurred, operation is complete
 	 1 = no error, but operation has not completed.
 	
 The LAPWrt code performs the following checks on the packet before writing
 it out:
 
 Check if LAP is up; if not, return noBridgeErr
 Check LAPtype; return LAPProtErr if it's <= 0
 Check the sums of the lengths of the WDS segments, return DDPLenErr if > 603
 If the packet was a short DDP, rewrite the header to be a long DDP
 if (TickCount() - tLastRx) > 30 seconds, nIM = TRUE
 Frame any LAP Control packets required
 Frame up the passed-in WDS
 PBWrite(the data, async);
 if noErr on PBWrite, return(1); otherwise return error status
 
 The ioCompletion of the serial driver's PBWrite calls the atlk resource's LAPWrtRtn
 code. This reloads the required registers and then jumps to the LAP Manager. This
 may kick off another LAPWrt call before returning.
 
*/

// register definitions for routines called by LAPWriteCode

#define rD0 (*regs).r[0]
#define rD2 (*regs).r[2]
#define rD3 (*regs).r[3]
#define rA0 (*regs).r[8]
#define rA1 (*regs).r[9]
#define rA2 (*regs).r[10]

OSErr	LAPWrt(tRegs *regs)
{
	char		c;
	WDSElement	*wp;
	Ptr		p;
	short		len;
	OSErr		osstat;
	short		numWDSEl;		// number of WDS elements
	short		lapT;			// lapType of the frame
	WDSElement	aWDS[12];		// a real WDS
	WDSElement	*theWDSp;		// the WDS used to write it out
	char		tempBuf[MAXLAPDATA+LAPFRAMING];	// to hold long DDP header
	short		i;
	long		pktLen;
			
	if (linkState != kLinkUp) 	// not up yet...
		return(noBridgeErr);
		
// check LAPType
	
	wp = (WDSElement*) rA1;			// wp points at WDS
	p = wp->entryPtr;			// p points at the data
	p += 2;					// point at LAPType
	lapT = *p;
	if (lapT <= 0) {
		histlog("LAPErr",lapT,0);
		return(lapProtErr);		// bogus LAP protocol type
	}
	
// check the length of the data to be sent

	len = 0;
	numWDSEl = 0;
	while (wp->entryLength != 0)	{	// loop thru the WDS elements
		len += wp->entryLength;
		wp++;				// next element
		numWDSEl++;			// and count the elements
	}
	if ((len <= 0) || (len > MAXLAPDATA+3))	{	// length error!
		DDebugStr("\pLength Error");		// DEBUG!
		histlog("DDPLenErr",len,0);
		return (ddpLenErr);
	}
		
// check for short DDP - if so, re-write the header & WDS to be a long DDP

	if ((lapT == kSDDP) && (numWDSEl <= 12)) { // enough 
		histlog("ReWrtDDP",rA1, 0);
		RewriteHeader((WDSElement *) rA1, &aWDS[0], (char *)&tempBuf);
		theWDSp = &aWDS[0];
		if ((len+8) > MAXLAPDATA+3) 		// DEBUG!
			DDebugStr("\pRewritten Hdr > MAXLAPLEN");
	}
	else theWDSp = (WDSElement *) rA1;
	
// check if we've heard from the other end recently: if not, queue an IM

	CheckRxTimer();
	
// queue up IM, UR, and data packets as required

	p = outBuf;				// start at the beginning
	pktLen = FrameAALAPCtls(p);		// queue any AALAP ctls required
	p += pktLen;
	pktLen += Frame(theWDSp, p);		// queue the data packet
	outFree = outBuf + pktLen;		// point at end of data 	

// send the packet(s) asynchronously

	osstat = SendOutBuf(&LAPWrtRtn); 	// start the output going
	if (osstat) 	{			
		numTxErrs++;			// remember trouble count
		histlog("LWrtErr",osstat,0);	// DEBUG!
	}
	else		{
		numDataTx++;			// update good frame stats
		osstat = 1;			// return pending status
	}
	histlog("LAPWrt  ", rA1, osstat);	// 
	return(osstat);
}

/*************
 * F R A M E *
 *************
  
 This code is called to move a packet into the current output buffer. It must
 perform the "escaping" required of an AALAP link: that is, when called upon to
 send the following characters, 
 
 	0x10, 0x11, 0x13, 0x91, 0x93, 0xA5
 
 the atlk will send the two character sequence below:
 
 	0x10 followed <original character XOR 0x40>
 	
 For example, 0x13 will be sent as these two characters: 0x10 0x53
 
 Frame returns the length the data actually added to the buffer.

*/

long	Frame(WDSElement* aWDS, char * theBuf)
{
	unsigned char	c;
	Ptr	p;
	short	len;
	char	*freeP;
	
	freeP = theBuf;					// start here
	outCRC = 0;					// clear out the CRC
	*freeP++ = kFrame;				// first, a framing char

	p = aWDS->entryPtr;				// Treat first segment 
	p++;						//  specially - don't 
	p++;						//  send s & d nodes
	len = aWDS->entryLength;			// make segment 2 bytes   
	len -= 2;					//  shorter as well

	do	{					// loop thru each segment
		while (len>0)			{	// do all chars in segment
			NextCRC(&outCRC,(unsigned char*) p); // accumulate CRC
			freeP = Escape((unsigned char*) p, freeP);  // escape the character
			p++;				// and move on
			len--;				// decr the length
		}
		aWDS++;					// bump to next segment
		p = aWDS->entryPtr;			// point at first char
		len = aWDS->entryLength;	
	}
	while (len != 0);				// end of segment loop

	c = outCRC & 0x00FF;				// tack on CRC
	freeP = Escape(&c, freeP);			// least sig. byte first
	c = outCRC >> 8;				// ms. byte next
	freeP = Escape(&c, freeP);		
	*freeP++ = kFrame;				// finally a closing kFrame
	BlockMove(&interFrame, freeP, IFSeqLen);	// add the IF sequence
	freeP += IFSeqLen;				// point at next free byte
	return(freeP - theBuf);
}

/***************
 * L A P E N Q *
 ***************
  
 This code will be called by the LAP manager to test a suggested network address.
 This value is specified by D2, and for ModemTalk, a non-extended ADEV, is an 
 8-bit node number. 
 
 On entry:
 	A0 = return address after the operation is complete
 	A1 -> WDS or portUse byte pointer (if sending probes)
 	A2 ->.MPP variables
 	D0 = selector for which kind of write operation to perform (byte)
 	D2 = AppleTalk destination address specifier.
 	D3 = port to write through
 
 This code need not send any packets: the AInstall code is responsible for 
 setting up the link and doing the IM/UR handshake with the other end. Once this
 procedure has been completed, the atlk connection has acquired a network 
 address. Thus, this code only needs to respond to the LAPEnq calls with a
 LSetInUse (or LNSetInUse) call to the LAP Manager if the address is incorrect.
 
*/

void	LAPEnq(tRegs *regs)
{
	register unsigned char	proposedNode;
	register Ptr		MPPp;
	OSErr	osstat;
	
	MPPp = (Ptr) rA2;
	proposedNode = rD2;				// the suggested node number
	
	if (propNode != proposedNode)	{		// bump number of useful
		numENQs++;				//	(different) ENQ calls
//		histlog("LAPEnq",proposedNode,hasAdrs); // DEBUG!
	}
	propNode = proposedNode;			// save in the global
	
	if (hasAdrs) 			{		// if we have a link adrs
		if (propNode != myNode)		{	// but bad suggestion
			osstat = LMSetInUse(MPPp);	// better guess again
			numNodeReject++;		// bump counter
		}
		else histlog("goodLEnq",propNode,numNodeReject);
	}
}

/*******************************
 * F R A M E A A L A P C T L S *
 *******************************
  
 This routine creates IM, UR, or DISC frames in the supplied buffer as required.
  
*/

long FrameAALAPCtls(char * theBuf)
{
	WDSElement	ctlWDS[2];		// WDS for an IM/UR/IF frame
	char		ctlData[10];		// data of the IM/UR/IF frame
	long		netNode, lmyNet, lmyNode, lmyRouter;		
	long		t;
	short		accumLen;
	char		*bufP;
	
	bufP = theBuf;
	lmyNet = myNet;
	lmyNode = myNode;
	lmyRouter = myRouter;
	if (nIM)			{	// queue an IM if required
		*bufP++ = kXOn;			// add Ctl-Q to shake remote free
		netNode = (lmyNet << 16) + (lmyNode << 8);
		BuildAAwds(ctlWDS, ctlData, kIM, 4, (Ptr) &netNode);
		accumLen = Frame(ctlWDS, bufP);
		bufP += accumLen; 
		numIMtx++;
		nIM = FALSE;
//		histlog("SndingIM",netNode,0);
	}
	
	if (nUR)			{ 	// queue a UR if need be
		netNode = (lmyNet << 16) + (lmyRouter << 8);
		BuildAAwds(ctlWDS, ctlData, kUR, 4, (Ptr) &netNode);
		accumLen = Frame(ctlWDS, bufP);
		bufP += accumLen; 
		numURtx++;
		nUR = FALSE;
//		histlog("SndingUR",netNode,0);
	}
	t = TickCount();
	if ((numDisc > 0) && (t - tNextDisc > 0)) { 	// need another kDISC
		BuildAAwds(ctlWDS, ctlData, kDISC, 1, (Ptr) &netNode);
		accumLen = Frame(ctlWDS, bufP);
		bufP += accumLen; 
		numDisc--;				// decr the count
		tNextDisc = t + HALFSECOND;		// remember next time
	}
	return (bufP - theBuf);				// return length
}

/***********************
 * B U I L D A A W D S *
 ***********************
  
 This builds a WDS to be used for an AALAP control frame (IM, UR, IF, DISC). It
 also fills in the buffer with the data passed in.
   
*/

void	BuildAAwds(WDSElement* aWDS, Ptr dataBuf, short lapT, 
				short theLen, Ptr theData)
{
	register Ptr	p;
	
	p = dataBuf + 2;			// move ahead two bytes (Frame() 
						//  ignores first two bytes)
	*p++ = lapT;				// set the LAP Type
	*p++ = *theData++;			// and up to three bytes of data
	*p++ = *theData++;
	*p++ = *theData++;
	
	aWDS->entryLength = theLen+2;		// fake the length for Frame()
	aWDS->entryPtr = dataBuf;		// point at the data
	aWDS++;
	aWDS->entryLength = 0;			// last entry in the WDS
}

/***********************
 * S E N D O U T B U F *
 ***********************
  
 This routine fills in the relevant fields of outPB and initiates the
 PBWrite routine. It returns any OSErr that comes back from the PBWrite.
 
 The caller passes in the address of a write completion routine, or NULL,
 to be called at the completion of the PBWrite routine.
  
*/

OSErr	SendOutBuf(ProcPtr theWriteCompl)
{
	long	l;
	OSErr	osstat;
	
	if (outPB.ioResult > 0) DDebugStr("\pDouble-Write, outPB");
	numPBWrites++;
	outPB.ioRefNum = outRefNum;
	outPB.ioCompletion = theWriteCompl;	
	outPB.ioBuffer = outBuf;
	outPB.ioReqCount = (outFree - outBuf);
	osstat = PBWriteAsync(&outPB);		// write out the data
	histlog ("SendBuf",(long)outBuf, (outPB.ioReqCount<<16) + osstat);
	return (osstat);		
}

/***************
 * E S C A P E *
 ***************
  
 Copy the character to the next position in destP, escaping it as 
 necessary. Returns the updated destination pointer.
  
*/

char	*Escape(unsigned char *p, char * destP)
{
	register char * d;
	
	d = destP;
	if (BitTst(magicChar, *p)) {		// if it needs an esc
		*d++ = kEsc;			// copy kESC
		*d++ = (*p++ ^ 0x40); 		// copy char XOR 0x40
	}
	else	*d++ = *p++;
	return (d);
}

/*****************************
 * R E W R I T E H E A D E R *
 *****************************
  
 Create a long DDP header from a short DDP header. Recopy the WDS into a temp
 WDS. Furthermore, expand the first segment (assuming it includes the entire
 short DDP header) into a temp buffer.
  
*/
void	RewriteHeader(WDSElement *wds1, WDSElement *wds2, char *buf)

{
	register char	*s,*d;		// pointers to source and dest buffer
	short		segLen;		// length of first segment
	char		dNode;		// destination (LAP) node
	register short	ddpLen;		// DDP length 
	short		myNetHi, myNetLo;
	short		i;
	
// copy the first segment, die if it's < 8 bytes (ie, doesn't include DDPType)

//	DDebugStr("\pRewriting! ");	// DEBUG!
	segLen = wds1->entryLength;
	if (segLen < 8)		{
		DDebugStr("\pShort first segment");
		wds2->entryLength = 0;	// make zero length pkt
		return;			// and bail out
	}
	
	// parse short header; collecting interesting fields
	
	s = wds1->entryPtr;		// get info about first segment
	dNode = *s++;			// get DDP Dest Node (from LAP dest node)
	s++; s++; 			// skip sNode, LAPType
	ddpLen = *s++;			// get high byte of DDPLen
	ddpLen = (ddpLen<<8) + *s++;	// add in low byte; s points at dSocket
	
	ddpLen += 8;			// long DDP hdr is 8 bytes longer
	myNetHi = (myNet >> 8);		// precompute these values
	myNetLo = myNet & 0x00FF;	
	
	// now fill in the new wds segment
	
	wds2->entryLength = segLen+8;	// new segment is 8 bytes longer
	wds2->entryPtr = buf;		// point at temp buffer
	d = buf;
	*d++ = myRouter;		// fill in LAP Dest (the router's adrs)
	*d++ = myNode;			// LAP Src
	*d++ = kLDDP;			// Long DDP LAPtype
	*d++ = (ddpLen>>8);		// high byte of (new) length
	*d++ = ddpLen & 0x00FF;		// and low byte of length
	*d++ = 0;			// no checksum
	*d++ = 0;
	*d++ = myNetHi;			// dest network
	*d++ = myNetLo;			
	*d++ = myNetHi;			// source network
	*d++ = myNetLo;			
	*d++ = dNode;			// dest node, from LAP header
	*d++ = myNode;			// source node
	
	for (i = segLen-5; i> 0; i--)	{ // finish copying first segment
		*d++ = *s++;		// 	always copy dSkt, sSkt, DDPType
	}
	
// copy the remaining segments

	while (TRUE)		{ 	// loop thru all segments
		wds1++;			// bump source and dest WDS pointers
		wds2++;
		if ((wds2->entryLength = wds1->entryLength) == 0) // got to end of WDS
			break;
		wds2->entryPtr = wds1->entryPtr;
	}

}

/***************************
 * C H E C K R X T I M E R *
 ***************************
  
 Check to see if we've received a frame recently. If not, force an IM
 to evoke a response.
  
*/

void	CheckRxTimer(void)
{
	long	t;
	
	t = TickCount();
	if (((t-tLastRx) > 30*SECOND) 		// if it has been a while
	 && (tStartedIM == 0)) 		{	// and we haven't asked recently
		tStartedIM = t;			
		nIM = TRUE;
	}
}