/******************************* BPrefsFile.c *********************************/ 
/*
	Copyright  1993, Trustees of Dartmouth College
	
	Richard E. Brown
	
	Release date:	29 March 1993 

 This object handles the interaction the preferences file. It allows you to
 read pref's from the file, and write them back easily. It also takes care of
 finding the "Preferences" folder, under System 6 or 7. (In System 7, it uses
 the Folder Manager as expected. In System 6, it'll look for a folder in the 
 system folder named "Preferences", creating one if needed.)
 
 Preferences are stored in STR# format. There may be an arbitrary number of
 preferences (up to 32,767...). Each preference is a Pascal string, and
 thus must be <= 255 characters long.
 
 This parrots much of Jamie McCarthy's CPrefsFile object, but removes the 
 Failxxx() machinery, because this is just a small code resource :-)
 
*/

#include <GestaltEqu.h>
#include <Folders.h>
#include "histlog.h"
#include <BPrefsFile.h>

#define PSLen(x) ((x) & 0x00FF)
#define debugPrefs 1

enum {
	kPFPrefsWereInvalidStrIndex = 1,
	kPFDefaultPrefsWereInvalidStrIndex,
	kPFCannotWriteStrIndex,
	kPFFileLockedStrIndex,
	kPFDiskLockedStrIndex,
	kPFFileBusyStrIndex,
	kPFUnknownErrorStrIndex,
	kPFPrefsFileNameStrIndex,
	kPFPrefsFolderNameStrIndex,
	kPFReserved,
	kPFFirstPrefsSuffixStrIndex
} ;

#define kPFEmptyStr ((unsigned char *) "\p")
#define kPFPrefsWereInvalidDefaultStr ((unsigned char *) \
	"\pThe preferences were invalid, and were replaced by the defaults.")
#define kPFDefaultPrefsWereInvalidDefaultStr ((unsigned char *) \
	"\pThe default preferences are invalid. This is a fatal error.")
#define kPFCannotWriteDefaultStr ((unsigned char *) \
	"\pYou wont be able to save preferences, because ")
#define kPFFileLockedDefaultStr ((unsigned char *) \
	"\pthe file is locked.")
#define kPFDiskLockedDefaultStr ((unsigned char *) \
	"\pthe disk is locked.")
#define kPFFileBusyDefaultStr ((unsigned char *) \
	"\pthe file is busy.")
#define kPFUnknownErrorDefaultStr ((unsigned char *) \
	"\pof an unexpected error.")
#define kPFPrefsFolderNameDefaultStr ((unsigned char *) "\pPreferences")
#define kPFLongPrefsSuffixDefaultStr ((unsigned char *) "\p Preferences")
#define kPFShortPrefsSuffixDefaultStr ((unsigned char *) "\p Prefs")

/***********************
 * I P R E F S F I L E *
 ***********************
 
 Initialize and open/create a preferences file. Pref's will be available
 after a successful call.
 
*/

OSErr 	BPrefsFile::IPrefsFile(OSType fileType, OSType sign,
			short strsID, ResType resTyp, short resID)
{
	OSErr	osstat;
	Str255	prefsFileName;
	
	itsFileType = fileType;			// file type 
	itsFileSign = sign;			// its signature
	itsStrsID = strsID;			// STR# of interesting strs
	itsRsrcType = resTyp;			// res type for the prefs
	itsRsrcID = resID;			// res ID for the prefs
	itsApplRefNum = CurResFile();		// get refnum of the current file...
	dirty = FALSE;				// we haven't changed it yet

	volNum = 0;
	dirID = 0;
	itsRefNum = 0;				// things may fail
	numStrs = 0;
	itsPrefRsrc = NULL;
	
	osstat = FindPrefsFolder();		// find/create prefs folder
	if (osstat) 	{
		if (debugPrefs) histlog("noPrefsFolder",osstat,0);
		osstat = LoadDefaultPrefs();	// get default prefs
		goto bailout;
	}
	
	HLock(this);
	GetIndString(prefsFileName, strsID, kPFPrefsFileNameStrIndex);
	
	if (!ExistsOnDisk(prefsFileName))	{
		osstat = LoadDefaultPrefs();	   // get default prefs
		if (osstat) goto bailout;
		osstat = CreateNewPrefs(prefsFileName, itsFileSign, itsFileType);
		if (osstat) goto bailout;
		osstat = TryToOpen(prefsFileName); // open it
		osstat = SavePrefs();		   // and save 'em
		if (osstat) return(osstat);
	}
	
	if (ExistsOnDisk(prefsFileName))	{
		osstat = TryToOpen(prefsFileName); // OK to fail 
	}
	HUnlock(this);
	
	osstat = LoadPrefs();			// read 'em in from prefs or adev
bailout:
	return (osstat);
}
		
/*****************
 * D I S P O S E *
 *****************
 
 Dispose of the pref's file. Save to disk if there are any changes, 
 then close the resource file.
 
*/

void	BPrefsFile::Dispose(void)
{
	OSErr	osstat;
	
	osstat = SavePrefs();
	if (itsRefNum) CloseResFile(itsRefNum);
	delete(this);
}

/*********************
 * L O A D P R E F S *
 *********************
 
 Get preferences from the prefs file.
 
*/

OSErr	BPrefsFile::LoadPrefs(void)
{
	short	oldResFile;
	OSErr	retval;
	short	*s;
	
	retval = noErr;
	if (itsRefNum) 	{
		oldResFile = CurResFile();
		UseResFile(itsRefNum);		// use the preferences file
		if (debugPrefs) histlog("loadPrefs",itsRefNum,oldResFile);
		itsPrefRsrc = Get1Resource(itsRsrcType, itsRsrcID);
		if (itsPrefRsrc) {
			s = (short *) *itsPrefRsrc;	/* s points at resource data */
			numStrs = *s;			/* get number of strings */
			if (debugPrefs) histlog("gotPrefs",numStrs,0);
		}
		else	retval = LoadDefaultPrefs();
		UseResFile(oldResFile);
	}
	else retval = LoadDefaultPrefs();	//   just use defaults
	return(retval);
}

/***********************************
 * L O A D D E F A U L T P R E F S *
 ***********************************
 
 Get default preferences from the main resource file, not the prefs file.
 
*/

OSErr	BPrefsFile::LoadDefaultPrefs(void)
{
	short	oldResFile;
	short	*s;

	oldResFile = CurResFile();
	UseResFile(itsApplRefNum);
	itsPrefRsrc = GetResource(itsRsrcType, itsRsrcID);
	if (!itsPrefRsrc) 	{
		itsPrefRsrc = NULL;
		numStrs = 0;
		if (debugPrefs) histlog("!!noDefault",0,0);
	}
	else	{
		DetachResource(itsPrefRsrc);	// detach it
		if (itsRefNum)		{	// if we have a prefs file...
			UseResFile(itsRefNum);	// point at prefs file
			AddResource(itsPrefRsrc, itsRsrcType, itsRsrcID,"\pPrefs");
			dirty = TRUE;
		}
		s = (short *) *itsPrefRsrc; 	/* s points at resource data */
		numStrs = *s;			/* get number of strings */
		if (debugPrefs) histlog("gotDefaults",itsRefNum,(long)itsPrefRsrc);
	}
	UseResFile(oldResFile);
	return(noErr);
}

/*********************
 * S A V E P R E F S *
 *********************
 
 Save the current preferences into the prefs file.
 
*/

OSErr	BPrefsFile::SavePrefs(void)
{
	short	oldResFile;
	
	if (dirty && itsRefNum)		{
		if (debugPrefs) histlog("SavingPrefs",0,0);
		oldResFile = CurResFile();
		UseResFile(itsRefNum);
		ChangedResource(itsPrefRsrc);	
		UpdateResFile(itsRefNum);	/* put it back on disk */
		UseResFile(oldResFile);			
		dirty = FALSE;	
	}
	return(noErr);
}

/*******************************
 * S A V E A S S E T T I N G S *
 *******************************
 
 Save the current settings in the specified file.
 
*/

OSErr	BPrefsFile::SaveAsSettings(short refNum)
{

}

/*******************
 * N U M P R E F S *
 *******************
 
 Return the number of preference strings in this resource.
 
*/

short	BPrefsFile::NumPrefs(void)
{
	return(numStrs);
}

/*******************
 * G E T A P R E F *
 *******************
 
 Return the preference from the resource handle. destLen is length of entire
 destination buffer, including the length byte.
 
*/

void	BPrefsFile::GetAPref(short n, char * thePref, long destLen)
{
 	char	* p;
 	short	* s;
 	short	i;
 	long	newLen;				// length of new setting string
 	
	s = (short *) *itsPrefRsrc;		/* s points at resource data */
	p = (char *) s + sizeof(short);		/* skip over the count of strings */
	if ((n >= 1) && (n <= numStrs)) {
		for (i=1; i<n; i++)		/* loop thru 'til we find it */
		{
			p += PSLen(*p) + 1;	/* add Pascal length to p to get next one */
		}
	}
	else p = (char *) &"\pNO SUCH SETTING!";
	
	// p now points to the start of the desired string...

	newLen = PSLen(*p)+1;		// length including length byte
	if (newLen > destLen) 
		newLen = destLen;	// trim if dest is too small
	BlockMove(p, thePref, newLen);	 // copy setting with length byte
}
	
/*******************
 * S E T A P R E F *
 *******************
 
 Update the n'th preference in the resource handle. If the n'th string
 has changed, then move the subsequent string(s) if needed, and copy in
 the new string.
 
*/

void	BPrefsFile::SetAPref(short n, char * thePref)
{
 	char	* p, *q;
 	short	* s;
 	short	i;
 	register long	rsrcLen, delta, copyLen, insOffset, oldLen, newLen;
 	
	s = (short *) *itsPrefRsrc;		/* s points count of strings */
	if ((n < 1) || (n > numStrs)) 		/* if a bogus string... */
		return;				/* just return */
		
	p = (char *) s+sizeof(short);		/* skip over the count of strings */
	for (i=1; i<n; i++)		/* loop thru 'til we find it */
	{
		p += PSLen(*p) + 1;	/* add Pascal length to p to get next one */
	}
		
// p now points to the start of the string to be updated. Now compute the 
// size of the move, and where to start the move (before we do anything 
// that moves memory)

	rsrcLen = GetHandleSize(itsPrefRsrc);	// get current resource length
 	q = *itsPrefRsrc;			// point at data
 	oldLen = PSLen(*p)+1;			// length of existing pref
 	newLen = PSLen(*thePref)+1;		// len. of new (incl length byte) 
	delta = newLen - oldLen;
	insOffset = p - q;			// insertion offset
	copyLen = (q + rsrcLen) - (p + oldLen);	// p to end of rsrc - len(p)
	
	if (!EqualString(p, thePref, FALSE, FALSE)) 		{		// something new?
		dirty = TRUE;
		if (delta > 0)			{	// if it'll be longer
			SetHandleSize(itsPrefRsrc, rsrcLen+delta); // bump handle size
			//if (MemError()) DDebugStr("\pSetHandleSize failed");
		}
	 	HLock(itsPrefRsrc);
	 	p = *itsPrefRsrc + insOffset;
		BlockMove(p+oldLen, p+newLen, copyLen);	// copy tail
		BlockMove(thePref, p, newLen);		// copy in new string
		HUnlock(itsPrefRsrc);
		if(delta < 0)			{	// it got shorter
			SetHandleSize(itsPrefRsrc, rsrcLen+delta);
			//if (MemError()) DDebugStr("\pSetHandleSize failed");
		}
	}
}

/*********************************
 * F I N D P R E F S F O L D E R *
 *********************************
 
 Find/Create a Preferences folder in the System folder. Sets volNum and 
 dirID instance variables to the right folder.
 
*/

OSErr	BPrefsFile::FindPrefsFolder(void)
{
	Boolean 	foundFolder;
	short 		myFoundVRefNum;
	long 		myFoundDirID;
	HParamBlockRec 	params;
	CInfoPBRec 	cInfo;
	SysEnvRec 	theWorld;
	OSErr		theOSErr;

	
	this->volNum = 0;
	this->dirID = 0;
	
	theOSErr = SysEnvirons(1, &theWorld);
	if (theOSErr == envNotPresent || theOSErr == envBadVers) {
		goto bailout;
	}
	/* if Sys 6, look manually in the Preferences folder */
	
	if (theWorld.systemVersion < 0x0700) 	{ 
		
		Str63 theFolder;
		
		if (debugPrefs) histlog("FFoldSy6",theWorld.systemVersion,0);
		GetIndString(theFolder, itsStrsID, kPFPrefsFolderNameStrIndex);
		
		params.volumeParam.ioCompletion = NULL;
		params.volumeParam.ioNamePtr = NULL;
		params.volumeParam.ioVRefNum = theWorld.sysVRefNum;
		params.volumeParam.ioVolIndex = 0;
		theOSErr = PBHGetVInfo(&params, FALSE);
		if (theOSErr) goto bailout;
		
		myFoundVRefNum = params.volumeParam.ioVRefNum;
		
			/* Look for the folder */
		cInfo.dirInfo.ioCompletion = NULL;
		cInfo.dirInfo.ioNamePtr = theFolder;
		cInfo.dirInfo.ioVRefNum = theWorld.sysVRefNum;
		cInfo.dirInfo.ioFDirIndex = 0;
		cInfo.dirInfo.ioDrDirID = 0;
		
		if (PBGetCatInfo(&cInfo, FALSE) == fnfErr) {
				/* There's no folder in the blessed folder called "Preferences"
				 * (or whatever we decided is its proper name).  Make one.
				 */
			theOSErr = DirCreate(theWorld.sysVRefNum, 0, theFolder, &myFoundDirID);
			if (theOSErr) goto bailout;
		} else {
			myFoundDirID = cInfo.dirInfo.ioDrDirID;
		}
		if (debugPrefs) histlog("Got6Folder",myFoundDirID,myFoundVRefNum);
	}

	else 	{ /* Use Gestalt to look in the Preferences folder */
		
		long foldAttr;
		
		if (debugPrefs) histlog("FFoldSy7",0,0);
		theOSErr = Gestalt(gestaltFindFolderAttr, &foldAttr);
		if ((theOSErr == noErr) && (foldAttr & (1L << gestaltFindFolderPresent))) {
			
			theOSErr = FindFolder(kOnSystemDisk, kPreferencesFolderType,
				TRUE, /* create the folder, if necessary */
				&myFoundVRefNum, &myFoundDirID);
				foundFolder = TRUE;
			if (theOSErr) goto bailout;
		}
	}
	

bailout:
	if (debugPrefs) histlog("FFoldResult",theOSErr,0);
	
	this->volNum = myFoundVRefNum;
	this->dirID = myFoundDirID;
	return (theOSErr);
}

/*******************************
 * C R E A T E N E W P R E F S *
 *******************************
 
 Create a new Preferences resource file in the given directory
 
*/

OSErr	BPrefsFile::CreateNewPrefs(Str63 filename, OSType creator, OSType fType)
{	
	OSErr	osstat;
	
	osstat = HCreate(volNum, dirID, filename, creator, fType);
	if (osstat) return (osstat);
 	HCreateResFile(volNum, dirID, filename);
 	osstat = ResError();
 	if (debugPrefs) histlog("CreatPref",osstat,0);
 	return (osstat);
}

/*********************
 * T R Y T O O P E N *
 *********************
 
 Open the Preferences resource file
 
*/

OSErr	BPrefsFile::TryToOpen(Str63 filename)
{
	OSErr	osstat;
	
	itsRefNum = HOpenResFile(volNum, dirID, filename, fsRdWrPerm);
	osstat = ResError();
	if (debugPrefs) histlog("TryToOpen",osstat,itsRefNum);
	return(osstat);	
}

/***************************
 * E X I S T S O N D I S K *
 ***************************
 
 Return true if the named file exists in the given directory
 
*/

Boolean BPrefsFile::ExistsOnDisk(Str63 fName)
{
	FInfo	fileInfo;
	OSErr	err;
	
	err = HGetFInfo( volNum, dirID, fName, &fileInfo);
	if (debugPrefs) histlog("existsOnDisk",err,0);
	return (err == noErr);
}